# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Command line interface.

Each command in the command line interface is implemented as a subclass of
:class:`Command`. Commands are also referenced from ``setup.py`` using the
entry point mechanism which allows the commands to be automatically
discovered.

The :func:`.main` function is the entry point of command line interface.
"""

from __future__ import division, unicode_literals

import re
import os
import sys
import argparse
import logging
import abc
import pickle
from itertools import islice
import multiprocessing as mp

import pkg_resources
from six import add_metaclass, iteritems, itervalues, text_type, PY3

from . import __version__ as package_version
from .datasource import native, sbml
from .datasource.context import FilePathContext
from .lpsolver import generic

import codecs

if PY3:
    unicode = str
if not PY3:
    if sys.stdout.encoding != 'UTF-8':
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout, 'strict')

logger = logging.getLogger(__name__)


def convert_to_unicode(str_, encoding='UTF-8'):
    if PY3 or bool(re.search(r'\\u', str_)):
        try:
            return str_.encode('latin-1').decode('unicode-escape')
        except:
            return str_
    else:
        if isinstance(str_, unicode):
            return str_
        return str_.decode(encoding)


class CommandError(Exception):
    """Error from running a command.

    This should be raised from a ``Command.run()`` if any arguments are
    misspecified. When the command is run and the ``CommandError`` is raised,
    the caller will exit with an error code and print appropriate usage
    information.
    """


@add_metaclass(abc.ABCMeta)
class Command(object):
    """Represents a command in the interface, operating on a model.

    The constructor will be given the NativeModel and the command line
    namespace. The subclass must implement :meth:`run` to handle command
    execution. The doc string will be used as documentation for the command
    in the command line interface.

    In addition, :meth:`init_parser` can be implemented as a classmethod which
    will allow the command to initialize an instance of
    :class:`argparse.ArgumentParser` as desired. The resulting argument
    namespace will be passed to the constructor.
    """

    def __init__(self, model, args):
        self._model = model
        self._args = args

        if self._model.name is not None:
            logger.info('Model: {}'.format(self._model.name))

        if self._model.version_string is not None:
            logger.info('Model version: {}'.format(self._model.version_string))

    @classmethod
    def init_parser(cls, parser):
        """Initialize command line parser (:class:`argparse.ArgumentParser`)"""

    @abc.abstractmethod
    def run(self):
        """Execute command"""

    def argument_error(self, msg):
        """Raise error indicating error parsing an argument."""
        raise CommandError(msg)

    def fail(self, msg, exc=None):
        """Exit command as a result of a failure."""
        logger.error(msg)
        if exc is not None:
            logger.debug('Command failure caused by exception!', exc_info=exc)
        sys.exit(1)


class MetabolicMixin(object):
    """Mixin for commands that use a metabolic model representation."""

    def __init__(self, *args, **kwargs):
        super(MetabolicMixin, self).__init__(*args, **kwargs)

        self._mm = self._model.create_metabolic_model()

    def report_flux_balance_error(self, exc=None):
        to_check = set()
        for reaction in self._mm.reactions:
            lower, upper = self._mm.limits[reaction]
            if upper < 0 or lower > 0:
                to_check.add(reaction)

        message = 'Failed to solve flux balance problem!'
        if len(to_check) > 0:
            message += ' Please check reactions with forced flux: {}'.format(
                ', '.join(sorted(to_check)))
        self.fail(message, exc)


class ObjectiveMixin(object):
    """Mixin for commands that use biomass as objective.

    Allows the user to override the default objective from the command line.
    """
    @classmethod
    def init_parser(cls, parser):
        parser.add_argument('--objective', type=convert_to_unicode,
                            help='Reaction to use as objective')
        super(ObjectiveMixin, cls).init_parser(parser)

    def _get_objective(self, log=True):
        if self._args.objective is not None:
            reaction = self._args.objective
        else:
            reaction = self._model.biomass_reaction
            if reaction is None:
                self.argument_error('The objective reaction was not specified')

        if log:
            logger.info('Using {} as objective'.format(reaction))

        return reaction


class LoopRemovalMixin(object):
    """Mixin for commands that perform loop removal."""

    _supported_loop_removal = ['none']

    __all_methods = {
        'none': 'Loop removal disabled; spurious loops are allowed',
        'tfba': 'Loop removal using thermodynamic constraints',
        'l1min': 'Loop removal using L1 minimization'
    }

    @classmethod
    def init_parser(cls, parser):
        supported = cls._supported_loop_removal
        assert len(supported) > 0, 'No loop removal methods defined'

        invalid = set(supported).difference(cls.__all_methods)
        assert len(invalid) == 0, 'Loop removal methods are invalid'

        if len(supported) >= 2:
            supported_list = ', '.join(s + ' (default)' if i == 0 else s
                                       for i, s in enumerate(supported))
            help_text = (
                'Select type of loop removal to perform.'
                ' Possible methods are: ' + supported_list)
            parser.add_argument(
                '--loop-removal', help=help_text, type=text_type,
                default=supported[0],
                metavar='METHOD', dest='hidden_mixin_loop_removal')

        super(LoopRemovalMixin, cls).init_parser(parser)

    def _get_loop_removal_option(self, log=True):
        supported = self._supported_loop_removal

        if len(supported) >= 2:
            loop_removal = self._args.hidden_mixin_loop_removal
        else:
            loop_removal = supported[0]

        if loop_removal not in self.__all_methods:
            self.argument_error(
                'Unknown loop removal method: {}'.format(loop_removal))

        if loop_removal not in supported:
            self.argument_error(
                'The loop removal method {} is not possible'
                ' with this command'.format(loop_removal))

        if log:
            logger.info(self.__all_methods[loop_removal])

        return loop_removal


class SolverCommandMixin(object):
    """Mixin for commands that use an LP solver.

    This adds a ``--solver`` parameter to the command that the user can use to
    select a specific solver. It also adds the method :meth:`_get_solver` which
    will return a solver with the specified default requirements. The user
    requirements will override the default requirements.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--solver', action='append', type=str,
            help='Specify solver requirements (e.g. "rational=yes")\n'
                 'Choices: \trational, integer, quadratic, '
                 'threads. feasibility_tolerance, optimality_tolerance, '
                 'integrality_tolerance')
        super(SolverCommandMixin, cls).init_parser(parser)

    def __init__(self, *args, **kwargs):
        super(SolverCommandMixin, self).__init__(*args, **kwargs)
        self._solver_args = {}
        if self._args.solver is not None:
            for s in self._args.solver:
                key, value = generic.parse_solver_setting(s)
                self._solver_args[key] = value

    def _get_solver(self, **kwargs):
        """Return a new :class:`psamm.lpsolver.lp.Solver` instance"""
        solver_args = dict(kwargs)
        solver_args.update(self._solver_args)
        return generic.Solver(**solver_args)


class ParallelTaskMixin(object):
    """Mixin for commands that run parallel computation tasks."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--parallel', help='Set number of parallel processes (0=auto)',
            type=int, default=0)
        super(ParallelTaskMixin, cls).init_parser(parser)

    def _create_executor(self, handler, args, cpus_per_worker=1):
        """Return a new :class:`.Executor` instance."""
        if self._args.parallel > 0:
            workers = self._args.parallel
        else:
            try:
                workers = mp.cpu_count() // cpus_per_worker
            except NotImplementedError:
                workers = 1

        if workers != 1:
            logger.info('Using {} parallel worker processes...'.format(
                workers))
            executor = ProcessPoolExecutor(
                processes=workers, handler_init=handler, handler_args=args)
        else:
            logger.info('Using single worker...')
            executor = SequentialExecutor(
                handler_init=handler, handler_args=args)

        return executor


class FilePrefixAppendAction(argparse.Action):
    """Action that appends one argument or multiple from a file.

    If the argument starts with a character in ``fromfile_prefix_chars``
    the remaining part of the argument is taken to be a file path. The file
    is read and every line is appended. Otherwise, the argument is simply
    appended.
    """
    def __init__(self, option_strings, dest, nargs=None,
                 fromfile_prefix_chars='@', **kwargs):
        if nargs is not None:
            raise ValueError('nargs not allowed')

        self.__fromfile_prefix_chars = fromfile_prefix_chars
        self.__final_type = kwargs.get('type')

        super(FilePrefixAppendAction, self).__init__(
            option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        arguments = getattr(namespace, self.dest)
        if arguments is None or len(arguments) == 0:
            arguments = []
            setattr(namespace, self.dest, arguments)
        values = text_type(values)
        if len(values) > 0 and values[0] in self.__fromfile_prefix_chars:
            filepath = values[1:]
            try:
                with open(filepath, 'r') as f:
                    for line in f:
                        arguments.append(self.__final_type(line.strip()))
            except IOError:
                parser.error('Unable to read arguments from file: {}'.format(
                    filepath))
        else:
            arguments.append(self.__final_type(values))


class _ErrorMarker(object):
    """Signals error in the child process."""

    def __init__(self, pickled_exc=None):
        self.pickled_exception = pickled_exc


class ExecutorError(Exception):
    """Error running tasks on executor."""


class _ExecutorProcess(mp.Process):
    def __init__(self, task_queue, result_queue, handler_init,
                 handler_args=()):
        super(_ExecutorProcess, self).__init__()
        self._task_queue = task_queue
        self._result_queue = result_queue
        self._handler_init = handler_init
        self._handler_args = handler_args

    def run(self):
        try:
            handler = self._handler_init(*self._handler_args)
            for tasks in iter(self._task_queue.get, None):
                results = [
                    (task, handler.handle_task(*task)) for task in tasks]
                self._result_queue.put(results)
        except BaseException as e:
            try:
                pickled_exc = pickle.dumps(e, -1)
            except Exception:
                logger.warning("Unpicklable exception raised: {}".format(e))
                pickled_exc = None
            self._result_queue.put(_ErrorMarker(pickled_exc))


class Executor(object):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        pass


class ProcessPoolExecutor(Executor):
    def __init__(self, handler_init, handler_args=(), processes=None):
        if processes is None:
            try:
                processes = mp.cpu_count()
            except NotImplementedError:
                processes = 1

        self._process_count = processes
        self._processes = []
        self._task_queue = mp.Queue()
        self._result_queue = mp.Queue()

        for _ in range(self._process_count):
            p = _ExecutorProcess(
                self._task_queue, self._result_queue, handler_init,
                handler_args)
            p.start()
            self._processes.append(p)

    def apply(self, task):
        self._task_queue.put([task])
        results = self._result_queue.get()
        if isinstance(results, _ErrorMarker):
            exception = pickle.loads(results.pickled_exception)
            raise exception

        return next(itervalues(results))

    def imap_unordered(self, iterable, chunksize=1):
        def iter_chunks():
            while True:
                chunk = list(islice(iterable, chunksize))
                if len(chunk) == 0:
                    break
                yield chunk

        it = iter_chunks()
        workers = 0
        for i in range(self._process_count):
            tasks = next(it, None)
            if tasks is None:
                break

            self._task_queue.put(tasks)
            workers += 1

        exception = None
        while workers > 0:
            results = self._result_queue.get()
            if isinstance(results, _ErrorMarker):
                if exception is None:
                    if results.pickled_exception is None:
                        exception = ExecutorError(
                            "Unpicklable exception raised by child")
                    else:
                        exception = pickle.loads(results.pickled_exception)
                workers -= 1
                self._process_count -= 1
                continue

            if exception is not None:
                continue

            tasks = next(it, None)
            if tasks is None:
                workers -= 1

            self._task_queue.put(tasks)

            for task, result in results:
                yield task, result

        if exception is not None:
            raise exception

    def close(self):
        for i in range(self._process_count):
            self._task_queue.put(None)

    def join(self):
        for p in self._processes:
            p.join()


class SequentialExecutor(Executor):
    def __init__(self, handler_init, handler_args=()):
        self._handler = handler_init(*handler_args)

    def apply(self, task):
        return self._handler.handle_task(*task)

    def imap_unordered(self, iterable, chunksize=1):
        for task in iterable:
            yield task, self._handler.handle_task(*task)

    def join(self):
        pass


def _trim(docstring):
    """Return a trimmed docstring.

    Code taken from 'PEP 257 -- Docstring Conventions' article.
    """
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = len(line) - len(stripped)
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    for line in lines[1:]:
        trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def main(command_class=None, args=None):
    """Run the command line interface with the given :class:`Command`.

    If no command class is specified the user will be able to select a specific
    command through the first command line argument. If the ``args`` are
    provided, these should be a list of strings that will be used instead of
    ``sys.argv[1:]``. This is mostly useful for testing.
    """

    # Set up logging for the command line interface
    if 'PSAMM_DEBUG' in os.environ:
        level = getattr(logging, os.environ['PSAMM_DEBUG'].upper(), None)
        if level is not None:
            logging.basicConfig(level=level)
    else:
        logging.basicConfig(level=logging.INFO)
        base_logger = logging.getLogger('psamm')
        if len(base_logger.handlers) == 0:
            handler = logging.StreamHandler()
            handler.setFormatter(
                logging.Formatter(u'%(levelname)s: %(message)s'))
            base_logger.addHandler(handler)
            base_logger.propagate = False

    title = 'Metabolic modeling tools'
    if command_class is not None:
        title, _, _ = command_class.__doc__.partition('\n\n')

    parser = argparse.ArgumentParser(description=title)
    parser.add_argument('--model', metavar='file', default='.',
                        help='Model definition')
    parser.add_argument(
        '-V', '--version', action='version',
        version='%(prog)s ' + package_version)

    if command_class is not None:
        # Command explicitly given, only allow that command
        command_class.init_parser(parser)
        parser.set_defaults(command=command_class)
    else:
        # Discover all available commands
        commands = {}
        for entry in pkg_resources.iter_entry_points('psamm.commands'):
            canonical = entry.name.lower()
            if canonical not in commands:
                command_class = entry.load()
                commands[canonical] = command_class
            else:
                logger.warning('Command {} was found more than once!'.format(
                    canonical))

        # Create parsers for subcommands
        subparsers = parser.add_subparsers(title='Commands', metavar='command')
        for name, command_class in sorted(iteritems(commands)):
            title, _, _ = command_class.__doc__.partition('\n\n')
            subparser = subparsers.add_parser(
                name, help=title.rstrip('.'),
                formatter_class=argparse.RawDescriptionHelpFormatter,
                description=_trim(command_class.__doc__))
            subparser.set_defaults(command=command_class)
            command_class.init_parser(subparser)

    parsed_args = parser.parse_args(args)

    # Load model definition
    model = native.ModelReader.reader_from_path(
        parsed_args.model).create_model()

    # Instantiate command with model and run
    command = parsed_args.command(model, parsed_args)
    try:
        command.run()
    except CommandError as e:
        parser.error(text_type(e))


def main_sbml(command_class=None, args=None):
    """Run the SBML command line interface."""
    # Set up logging for the command line interface
    if 'PSAMM_DEBUG' in os.environ:
        level = getattr(logging, os.environ['PSAMM_DEBUG'].upper(), None)
        if level is not None:
            logging.basicConfig(level=level)
    else:
        logging.basicConfig(level=logging.INFO)
        base_logger = logging.getLogger('psamm')
        if len(base_logger.handlers) == 0:
            handler = logging.StreamHandler()
            handler.setFormatter(
                logging.Formatter(u'%(levelname)s: %(message)s'))
            base_logger.addHandler(handler)
            base_logger.propagate = False

    title = 'Metabolic modeling tools (SBML)'
    if command_class is not None:
        title, _, _ = command_class.__doc__.partition('\n\n')

    parser = argparse.ArgumentParser(description=title)
    parser.add_argument('model', metavar='file', help='SBML file')
    parser.add_argument('--merge-compounds', action='store_true',
                        help=('Merge identical compounds occuring in various'
                              ' compartments.'))
    parser.add_argument(
        '-V', '--version', action='version',
        version='%(prog)s ' + package_version)

    if command_class is not None:
        # Command explicitly given, only allow that command
        command_class.init_parser(parser)
        parser.set_defaults(command=command_class)
    else:
        # Discover all available commands
        commands = {}
        for entry in pkg_resources.iter_entry_points('psamm.commands'):
            canonical = entry.name.lower()
            if canonical not in commands:
                command_class = entry.load()
                commands[canonical] = command_class
            else:
                logger.warning('Command {} was found more than once!'.format(
                    canonical))

        # Create parsers for subcommands
        subparsers = parser.add_subparsers(title='Commands', metavar='command')
        for name, command_class in sorted(iteritems(commands)):
            title, _, _ = command_class.__doc__.partition('\n\n')
            subparser = subparsers.add_parser(
                name, help=title.rstrip('.'),
                formatter_class=argparse.RawDescriptionHelpFormatter,
                description=_trim(command_class.__doc__))
            subparser.set_defaults(command=command_class)
            command_class.init_parser(subparser)

    parsed_args = parser.parse_args(args)

    # Load model definition
    context = FilePathContext(parsed_args.model)
    with context.open('r') as f:
        model = sbml.SBMLReader(f, context=context).create_model()
        sbml.convert_sbml_model(model)
        if parsed_args.merge_compounds:
            sbml.merge_equivalent_compounds(model)

    # Instantiate command with model and run
    command = parsed_args.command(model, parsed_args)
    try:
        command.run()
    except CommandError as e:
        parser.error(text_type(e))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>

"""Reaction parser."""
from __future__ import unicode_literals

import re
from decimal import Decimal
import enum

from six import raise_from, PY3

from psamm.reaction import Reaction, Compound, Direction
from psamm.expression import affine

if PY3:
    unicode = str


def convert_to_unicode(str_, encoding='UTF-8'):
    if PY3 or bool(re.search(r'\\u', str_)):
        try:
            return str_.encode('latin-1').decode('unicode-escape')
        except:
            return str_
    else:
        if isinstance(str_, unicode):
            return str_
        return str_.decode(encoding)


class ParseError(Exception):
    """Error raised when parsing reaction fails."""

    def __init__(self, *args, **kwargs):
        self._span = kwargs.pop('span', None)
        super(ParseError, self).__init__(*args, **kwargs)

    @property
    def indicator(self):
        if self._span is None:
            return None
        pre = ' ' * self._span[0]
        ind = '^' * max(1, self._span[1] - self._span[0])
        return pre + ind


@enum.unique
class _ReactionToken(enum.Enum):
    Space = 0
    Quoted = 1
    Group = 2
    Plus = 3
    Arrow = 4
    Other = 5
    End = 6


class ReactionParser(object):
    """Parser of reactions equations.

    This parser recognizes:

    * Global compartment specification as a prefix (when ``parse_global`` is
      ``True``)

    * Configurable reaction arrow tokens (``arrows``)

    * Compounds quoted by pipe (``|``) (required only if the compound name
      includes a space)

    * Compound counts that are affine expressions.
    """

    def __init__(self, arrows=None, parse_global=False):
        if arrows is None:
            arrows = (
                ('<=>', Direction.Both),
                ('=>', Direction.Forward),
                ('<=', Direction.Reverse)
            )

        arrow_p = '|'.join(re.escape(arrow) for arrow, _ in arrows)
        self._arrows = dict(arrows)
        self._tokens = (
            (r'\s+', _ReactionToken.Space),
            (r'\|[^|]+\|', _ReactionToken.Quoted),
            (r'\([^)]+\)(?=\Z|\s|\|)', _ReactionToken.Group),
            (r'\+(?=\s)', _ReactionToken.Plus),
            (arrow_p + r'(?=\Z|\s)', _ReactionToken.Arrow),
            (r'\S+', _ReactionToken.Other),
            (r'\Z', _ReactionToken.End),
            (r'.', None)
        )

        self._scanner = re.compile(
            '|'.join('(' + pattern + ')' for pattern, _ in self._tokens),
            re.DOTALL | re.UNICODE)

        self._parse_global = parse_global

    def parse(self, s):
        """Parse reaction string."""
        global_comp = None
        if self._parse_global:
            # Split by colon for global compartment information
            m = re.match(r'^\s*\[(\w+)\]\s*:\s*(.*)', s)
            if m is not None:
                global_comp = m.group(1)
                s = m.group(2)
        expect_operator = False
        direction = None
        left = []
        right = []
        current_side = left
        saved_token = None

        def tokenize():
            for match in re.finditer(self._scanner, s):
                for i, group in enumerate(match.groups()):
                    if group is not None:
                        token = self._tokens[i][1]
                        span = match.start(), match.end()
                        if token is None:
                            raise ParseError(
                                'Invalid token in expression string:'
                                ' {!r}'.format(group), span=span)
                        yield self._tokens[i][1], group, span
                        break

        for token, value, span in tokenize():
            # Handle partially parsed compound.
            if saved_token is not None and (
                    token in (_ReactionToken.Plus,
                              _ReactionToken.Arrow,
                              _ReactionToken.End)):
                compound = parse_compound(saved_token, global_comp)
                current_side.append((compound, 1))
                expect_operator = True
                saved_token = None

            if token == _ReactionToken.Plus:
                # Compound separator. Expect compound name or compound count
                # next.
                if not expect_operator:
                    raise ParseError('Unexpected token: {!r}'.format(value),
                                     span=span)
                expect_operator = False
            elif token == _ReactionToken.Arrow:
                # Reaction arrow. Expect compound name or compound count next.
                if direction is not None:
                    raise ParseError(
                        'More than one equation arrow: {!r}'.format(value),
                        span=span)

                if not expect_operator and len(left) > 0:
                    raise ParseError('Unexpected token: {!r}'.format(value),
                                     span=span)

                expect_operator = False
                direction = self._arrows[value]
                current_side = right
            elif token == _ReactionToken.Group:
                # Compound count. Expect compound name next.
                if expect_operator:
                    raise ParseError('Expected plus or arrow: {!r}'.format(
                        value), span=span)
                if saved_token is not None:
                    raise ParseError('Expected compound name: {!r}'.format(
                        value), span=span)
                saved_token = value
            elif token == _ReactionToken.Quoted:
                # Compound name. Expect operator next.
                if expect_operator:
                    raise ParseError('Expected plus or arrow: {!r}'.format(
                        value), span=span)
                if saved_token is not None:
                    try:
                        count = parse_compound_count(saved_token)
                    except ValueError as e:
                        raise_from(ParseError(
                            'Unable to parse compound count: {!r}'.format(
                                saved_token),
                            span=span), e)
                else:
                    count = 1
                compound = parse_compound(value, global_comp)
                current_side.append((compound, count))
                expect_operator = True
                saved_token = None
            elif token == _ReactionToken.Other:
                # Could be count or compound name. Store and expect other,
                # quoted, operator or end.
                if expect_operator:
                    raise ParseError('Expected plus or arrow: {!r}'.format(
                        value), span=span)
                if saved_token is not None:
                    try:
                        count = parse_compound_count(saved_token)
                    except ValueError as e:
                        raise_from(ParseError(
                            'Unable to parse compound count: {!r}'.format(
                                saved_token),
                            span=span), e)
                    compound = parse_compound(value, global_comp)
                    current_side.append((compound, count))
                    expect_operator = True
                    saved_token = None
                else:
                    saved_token = value

        if direction is None:
            raise ParseError('Missing equation arrow')
        return Reaction(direction, left, right)


_DEFAULT_PARSER = ReactionParser()


def parse_reaction(s):
    """Parse reaction string using the default parser."""
    return _DEFAULT_PARSER.parse(s)


def parse_compound(s, global_compartment=None):
    """Parse a compound specification.

    If no compartment is specified in the string, the global compartment
    will be used.
    """
    m = re.match(r'^\|(.*)\|$', s)
    if m:
        s = m.group(1)

    m = re.match(r'^(.+)\[(\S+)\]$', s)
    if m:
        compound_id = convert_to_unicode(m.group(1))
        compartment = m.group(2)
    else:
        compound_id = convert_to_unicode(s)
        compartment = global_compartment
    return Compound(compound_id, compartment=compartment)


def parse_compound_count(s):
    """Parse a compound count (number of compounds)."""
    m = re.match(r'^\((.*)\)$', s)
    if m:
        s = convert_to_unicode(m.group(1))

    for count_type in (int, Decimal, affine.Expression):
        try:
            return count_type(s)
        except:
            pass

    raise ValueError('Unable to parse compound count: {}'.format(s))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Implementation of fastGapFill.

Described in [Thiele14]_.
"""

from __future__ import unicode_literals

import logging

from .fastcore import fastcore

logger = logging.getLogger(__name__)


def fastgapfill(model_extended, core, solver, weights={}, epsilon=1e-5):
    """Run FastGapFill gap-filling algorithm by calling
    :func:`psamm.fastcore.fastcore`.

    FastGapFill will try to find a minimum subset of reactions that includes
    the core reactions and it also has no blocked reactions.
    Return the set of reactions in the minimum subset. An extended model that
    includes artificial transport and exchange reactions can be generated by
    calling :func:`.create_extended_model`.

    Args:
        model: :class:`psamm.metabolicmodel.MetabolicModel`.
        core: reactions in the original metabolic model.
        weights: a weight dictionary for reactions in the model.
        solver: linear programming library to use.
        epsilon: float number, threshold for Fastcore algorithm.
    """

    # Run Fastcore and print the induced reaction set
    logger.info('Calculating Fastcore induced set on model')
    induced = fastcore(
        model_extended, core, epsilon=epsilon, weights=weights, solver=solver)
    logger.debug('Result: |A| = {}, A = {}'.format(len(induced), induced))
    added_reactions = induced - core
    logger.debug('Extended: |E| = {}, E = {}'.format(
        len(added_reactions), added_reactions))
    return induced



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import os
import shutil
import tempfile
import unittest
import math
from decimal import Decimal
from collections import OrderedDict

from psamm.datasource import native, context, entry
from psamm.reaction import Reaction, Compound, Direction
from psamm.formula import Formula

from six import StringIO
import yaml


class MockEntry(object):
    """Entry object for testing OrderedEntrySet."""
    def __init__(self, id):
        self._id = id

    @property
    def id(self):
        return self._id


class TestOrderedEntrySet(unittest.TestCase):
    def test_init_from_dict(self):
        entries = {
            'test_1': MockEntry('test_1'),
            'test_2': MockEntry('test_2')
        }
        entry_set = native._OrderedEntrySet(entries)
        self.assertEqual(entry_set['test_1'], entries['test_1'])
        self.assertEqual(entry_set['test_2'], entries['test_2'])
        self.assertEqual(len(entry_set), len(entries))

    def test_add_and_contains(self):
        entry_set = native._OrderedEntrySet()
        entry_set.add_entry(MockEntry('def'))
        entry_set.add_entry(MockEntry('abc'))
        entry_set.add_entry(MockEntry('ghi'))

        self.assertIn('def', entry_set)
        self.assertIn('ghi', entry_set)
        self.assertNotIn('klm', entry_set)
        self.assertEqual(len(entry_set), 3)

        # Check that it is ordered
        self.assertEqual(
            [entry.id for entry in entry_set], ['def', 'abc', 'ghi'])

    def test_get(self):
        entries = {
            'test_1': MockEntry('test_1'),
            'test_2': MockEntry('test_2')
        }
        entry_set = native._OrderedEntrySet(entries)
        self.assertEqual(entry_set.get('test_1'), entries['test_1'])
        self.assertIsNone(entry_set.get('test_3'))
        self.assertEqual(entry_set.get('test_3', 5), 5)

    def test_discard(self):
        entries = {
            'test_1': MockEntry('test_1'),
            'test_2': MockEntry('test_2')
        }
        entry_set = native._OrderedEntrySet(entries)
        entry_set.discard('test_1')
        self.assertNotIn('test_1', entry_set)

    def test_update(self):
        entries = {
            'test_1': MockEntry('test_1'),
            'test_2': MockEntry('test_2')
        }
        entry_set = native._OrderedEntrySet(entries)
        entry_set.update([
            MockEntry('test_2'),
            MockEntry('test_3')
        ])

        self.assertEqual(len(entry_set), 3)
        self.assertEqual(entry_set['test_1'], entries['test_1'])
        self.assertNotEqual(entry_set['test_2'], entries['test_2'])

    def test_clear(self):
        entries = {
            'test_1': MockEntry('test_1'),
            'test_2': MockEntry('test_2')
        }
        entry_set = native._OrderedEntrySet(entries)
        entry_set.clear()

        self.assertEqual(len(entry_set), 0)
        self.assertNotIn('test_1', entry_set)


class TestYAMLDataSource(unittest.TestCase):
    def test_parse_compartments(self):
        model_dict = {
            'compartments': [
                {
                    'id': 'e',
                    'name': 'Extracellular'
                },
                {
                    'id': 'p',
                    'name': 'Periplasm',
                    'adjacent_to': ['e', 'c']
                },
                {
                    'id': 'c',
                    'name': 'Cytosol',
                    'adjacent_to': 'p'
                },
                {
                    'id': 'internal',
                    'name': 'Internal compartment',
                    'adjacent_to': 'c'
                }
            ]
        }
        reader = native.ModelReader(model_dict)
        compartment_iter, boundaries = reader.parse_compartments()
        compartments = list(compartment_iter)

        self.assertEqual(len(compartments), 4)
        self.assertEqual(compartments[0].id, 'e')
        self.assertEqual(compartments[0].name, 'Extracellular')
        self.assertEqual(compartments[1].id, 'p')
        self.assertEqual(compartments[1].name, 'Periplasm')
        self.assertEqual(compartments[2].id, 'c')
        self.assertEqual(compartments[2].name, 'Cytosol')
        self.assertEqual(compartments[3].id, 'internal')
        self.assertEqual(compartments[3].name, 'Internal compartment')

        normalized_boundaries = set(
            tuple(sorted((c1, c2))) for c1, c2 in boundaries)
        self.assertSetEqual(normalized_boundaries, {
            ('c', 'internal'),
            ('c', 'p'),
            ('e', 'p')
        })

    def test_parse_reaction(self):
        reaction = native.parse_reaction({
            'id': 'reaction_123',
            'equation': 'A + 2 B => C[e]'
        }, default_compartment='p')

        self.assertEqual(reaction.id, 'reaction_123')
        self.assertEqual(reaction.equation, Reaction(
            Direction.Forward,
            [(Compound('A', 'p'), 1), (Compound('B', 'p'), 2)],
            [(Compound('C', 'e'), 1)]))

    def test_parse_reaction_list(self):
        reactions = list(native.parse_reaction_list('./test.yaml', [
            {
                'id': 'rxn1',
                'equation': {
                    'reversible': True,
                    'left': [
                        {'id': 'A', 'value': 1},
                        {'id': 'B', 'value': 2}],
                    'right': [
                        {'id': 'C', 'value': 1}
                    ]
                }
            }
        ]))

        self.assertEqual(len(reactions), 1)

        reaction = Reaction(Direction.Both,
                            [(Compound('A'), 1), (Compound('B'), 2)],
                            [(Compound('C'), 1)])
        self.assertEqual(reactions[0].equation, reaction)

    def test_parse_reaction_list_missing_value(self):
        with self.assertRaises(native.ParseError):
            reaction = list(native.parse_reaction_list('./test.yaml', [
                {
                    'id': 'rxn1',
                    'equation': {
                        'left': [
                            {'id': 'A'}
                        ]
                    }
                }
            ]))

    def test_parse_exchange_table(self):
        table = '''
ac      e
glcD    e       -10
co2     e       -       50
'''

        exchange = list(
            native.parse_exchange_table_file(StringIO(table.strip())))
        self.assertEqual(len(exchange), 3)
        self.assertEqual(exchange[0], (Compound('ac', 'e'), None, None, None))
        self.assertEqual(exchange[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(exchange[2], (Compound('co2', 'e'), None, None, 50))

    def test_parse_exchange(self):
        exchange = list(native.parse_exchange({
            'compartment': 'e',
            'compounds': [
                {'id': 'ac'},
                {'id': 'glcD', 'lower': -10},
                {'id': 'co2', 'upper': 50},
                {'id': 'compound_x', 'compartment': 'c'},
                {'id': 'compound_y', 'reaction': 'EX_cpdy'}
            ]
        }, 'e'))

        self.assertEqual(len(exchange), 5)
        self.assertEqual(exchange[0], (Compound(('ac'), ('e')),
                                       None, None, None))
        self.assertEqual(exchange[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(exchange[2], (Compound('co2', 'e'), None, None, 50))
        self.assertEqual(
            exchange[3], (Compound('compound_x', 'c'), None, None, None))
        self.assertEqual(
            exchange[4], (Compound('compound_y', 'e'), 'EX_cpdy', None, None))

    def test_parse_normal_float(self):
        v = native.yaml_load('-23.456')
        self.assertEqual(v, Decimal('-23.456'))
        self.assertIsInstance(v, Decimal)

    def test_parse_special_float(self):
        self.assertEqual(native.yaml_load('.inf'), Decimal('Infinity'))
        self.assertEqual(native.yaml_load('-.inf'), -Decimal('Infinity'))
        self.assertTrue(math.isnan(native.yaml_load('.nan')))


class TestYAMLFileSystemData(unittest.TestCase):
    """Test loading files from file system."""

    def setUp(self):
        self._model_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def write_model_file(self, filename, contents):
        path = os.path.join(self._model_dir, filename)
        with open(path, 'w') as f:
            f.write(contents)
        return path

    def test_long_string_model(self):
        long_string = '''---
            name: Test model
            biomass: rxn_1
            reactions:
              - id: rxn_1
                equation: '|A_\u2206[e]| => |B[c]|'
                genes:
                  - gene_1
                  - gene_2
              - id: rxn_2_\u03c0
                equation: '|B[c]| => |C[e]|'
                genes: 'gene_3 or (gene_4 and gene_5)'
            compounds:
              - id: A_\u2206
              - id: B
              - id: C
            exchange:
              - compartment: e
                compounds:
                  - id: A_\u2206
                  - id: C
            limits:
              - reaction: rxn_2_\u03c0
                upper: 100
            '''
        m = native.ModelReader(long_string)
        self.assertEqual(m.name, 'Test model')
        self.assertEqual(m.biomass_reaction, 'rxn_1')
        self.assertEqual(m.extracellular_compartment, 'e')
        reactions = list(m.parse_reactions())
        self.assertEqual(reactions[0].id, 'rxn_1')

    def test_long_string_model_include(self):
        longString = '''---
            name: Test model
            biomass: rxn_1
            reactions:
              - include: exchange.yaml
            '''
        m = native.ModelReader(longString)
        with self.assertRaises(context.ContextError):
            list(m.parse_reactions())

    def test_dict_model(self):
        dict_model = {
            'name': 'Test model',
            'biomass': 'rxn_1',
            'reactions': [
                {'id': 'rxn_1',
                 'equation': '|A_\u2206[e]| => |B[c]|',
                 'genes': [
                    'gene_1',
                    'gene_2']
                 },
                {'id': 'rxn_2_\u03c0',
                 'equation': '|B[c]| => |C[e]|',
                 'genes': 'gene_3 or (gene_4 and gene_5)'
                 }],
            'compounds': [
                {'id': 'A_\u2206'},
                {'id': 'B'},
                {'id': 'C'}],
            'exchange': [
                {'compartment': 'e',
                 'compounds': [
                     {'id': 'A_\u2206'},
                     {'id': 'C'}]
                 }],
            'limits': [
                {'reaction': 'rxn_2_\u03c0',
                 'upper': 100
                 }]
        }

        dmodel = native.ModelReader(dict_model)
        self.assertEqual(dmodel.name, 'Test model')
        self.assertEqual(dmodel.biomass_reaction, 'rxn_1')
        self.assertEqual(dmodel.extracellular_compartment, 'e')

    def test_parse_model_file(self):
        path = self.write_model_file('model.yaml', '\n'.join([
            'name: Test model',
            'biomass: biomass_reaction_id',
            'extracellular: Extra',
            'default_flux_limit: 500'
        ]))

        model = native.ModelReader.reader_from_path(path)

        self.assertEqual(model.name, 'Test model')
        self.assertEqual(model.biomass_reaction, 'biomass_reaction_id')
        self.assertEqual(model.extracellular_compartment, 'Extra')
        self.assertEqual(model.default_flux_limit, 500)

    def test_bad_path(self):
        with self.assertRaises(native.ParseError):
            native.ModelReader.reader_from_path('/nope/nreal/path')

    def test_invalid_model_type(self):
        with self.assertRaises(ValueError):
            native.ModelReader(42.2)

    def test_parse_model_file_with_exchange(self):
        path = self.write_model_file('model.yaml', '\n'.join([
            'extracellular: Ex',
            'exchange:',
            ' - compounds:',
            '    - id: A',
            '    - id: B',
            '      compartment: c'
        ]))

        model = native.ModelReader.reader_from_path(path)

        exchange = list(model.parse_exchange())
        self.assertEqual(exchange[0][0], Compound('A', 'Ex'))
        self.assertEqual(exchange[1][0], Compound('B', 'c'))

    def test_parse_model_file_with_media(self):
        """Test parsing model with the deprecated media key."""
        path = self.write_model_file('model.yaml', '\n'.join([
            'extracellular: Ex',
            'media:',
            ' - compounds:',
            '    - id: A',
            '    - id: B',
            '      compartment: c'
        ]))

        model = native.ModelReader.reader_from_path(path)

        exchange = list(model.parse_exchange())
        self.assertEqual(exchange[0][0], Compound('A', 'Ex'))
        self.assertEqual(exchange[1][0], Compound('B', 'c'))

    def test_parse_model_file_with_media_and_exchange(self):
        """Test that parsing model with both media and exchange fails."""
        path = self.write_model_file('model.yaml', '\n'.join([
            'extracellular: Ex',
            'media:',
            ' - compounds:',
            '    - id: A',
            '    - id: B',
            '      compartment: c',
            'exchange:',
            ' - compounds:',
            '    - id: C',
            '    - id: D'
        ]))

        model = native.ModelReader.reader_from_path(path)

        with self.assertRaises(native.ParseError):
            medium = list(model.parse_exchange())

    def test_parse_compound_tsv_file(self):
        path = self.write_model_file('compounds.tsv', '\n'.join([
            'id\tname\tformula\tcharge\tkegg\tcas',
            'h2o\tWater\tH2O\t0\tC00001\t7732-18-5'
        ]))

        compounds = list(native.parse_compound_file(path, 'tsv'))
        self.assertEqual(len(compounds), 1)

        self.assertEqual(compounds[0].id, 'h2o')
        self.assertEqual(compounds[0].name, 'Water')
        self.assertEqual(compounds[0].properties['name'], 'Water')
        self.assertEqual(compounds[0].formula, 'H2O')
        self.assertEqual(compounds[0].charge, 0)
        self.assertEqual(compounds[0].properties['kegg'], 'C00001')
        self.assertEqual(compounds[0].properties['cas'], '7732-18-5')

    def test_parse_compound_yaml_file(self):
        path = self.write_model_file('compounds.yaml', '\n'.join([
            '- id: h2o',
            '  name: Water',
            '  formula: H2O',
            '  user_annotation: ABC'
        ]))

        compounds = list(native.parse_compound_file(path, 'yaml'))
        self.assertEqual(len(compounds), 1)

        self.assertEqual(compounds[0].id, 'h2o')
        self.assertEqual(compounds[0].name, 'Water')
        self.assertEqual(compounds[0].formula, 'H2O')
        self.assertEqual(compounds[0].properties['user_annotation'], 'ABC')

    def test_parse_reaction_table_file(self):
        path = self.write_model_file('reactions.tsv', '\n'.join([
            'id\tname\tequation',
            'rxn_1\tReaction 1\t|A| => (2) |B|',
            'rxn_2\tSecond reaction\t<=> |C|'
        ]))

        reactions = list(native.parse_reaction_file(path))
        self.assertEqual(len(reactions), 2)

        self.assertEqual(reactions[0].id, 'rxn_1')
        self.assertEqual(reactions[0].name, 'Reaction 1')
        self.assertEqual(reactions[0].properties['name'], 'Reaction 1')
        self.assertEqual(reactions[0].equation, Reaction(
            Direction.Forward, [(Compound('A'), 1)], [(Compound('B'), 2)]))
        self.assertEqual(reactions[0].properties.get('ec'), None)
        self.assertEqual(reactions[0].genes, None)
        self.assertEqual(reactions[0].filemark.filecontext.filepath, path)
        self.assertEqual(reactions[0].filemark.line, 2)

        self.assertEqual(reactions[1].id, 'rxn_2')
        self.assertEqual(reactions[1].name, 'Second reaction')
        self.assertEqual(reactions[1].equation, Reaction(
            Direction.Both, [], [(Compound('C'), 1)]))
        self.assertEqual(reactions[1].filemark.filecontext.filepath, path)
        self.assertEqual(reactions[1].filemark.line, 3)

    def test_parse_reaction_yaml_file(self):
        path = self.write_model_file('reaction.yaml', '\n'.join([
            '- id: rxn_1',
            '  equation: "|A| => |B|"',
            '- id: rxn_2',
            '  equation:',
            '    reversible: no',
            '    left:',
            '      - id: B',
            '        value: 1',
            '    right:',
            '      - id: A',
            '        value: 1'
        ]))

        reactions = list(native.parse_reaction_file(path))
        self.assertEqual(len(reactions), 2)

        self.assertEqual(reactions[0].id, 'rxn_1')
        self.assertEqual(reactions[1].id, 'rxn_2')

    def test_parse_exchange_table_file(self):
        path = self.write_model_file('exchange.tsv', '\n'.join([
            '',
            '# comment',
            'cpd_A\tc',
            'cpd_B\te\t-1000',
            'cpd_C\te\t-\t20  # line comment',
            'cpd_D\te\t-100\t-10'
        ]))

        exchange = list(native.parse_exchange_file(path, 'e'))
        self.assertEqual(exchange, [
            (Compound('cpd_A', 'c'), None, None, None),
            (Compound('cpd_B', 'e'), None, -1000, None),
            (Compound('cpd_C', 'e'), None, None, 20),
            (Compound('cpd_D', 'e'), None, -100, -10)
        ])

    def test_get_limits_invalid_fixed(self):
        d = {
            'fixed': 10,
            'upper': 20
        }
        with self.assertRaises(native.ParseError):
            native.get_limits(d)

    def test_parse_exchange_yaml_file(self):
        path = self.write_model_file('exchange.yaml', '\n'.join([
            'compartment: e',
            'compounds:',
            '  - id: cpd_A',
            '    reaction: EX_A',
            '    lower: -40',
            '  - id: cpd_B',
            '    upper: 100',
            '  - id: cpd_C',
            '    lower: -100.0',
            '    upper: 500.0',
            '  - id: cpd_D',
            '    compartment: c',
            '  - id: cpd_E',
            '    fixed: 100.0',
        ]))

        exchange = list(native.parse_exchange_file(path, 'e'))
        self.assertEqual(exchange, [
            (Compound('cpd_A', 'e'), 'EX_A', -40, None),
            (Compound('cpd_B', 'e'), None, None, 100),
            (Compound('cpd_C', 'e'), None, -100, 500),
            (Compound('cpd_D', 'c'), None, None, None),
            (Compound('cpd_E', 'e'), None, 100, 100)
        ])

    def test_parse_exchange_yaml_list(self):
        self.write_model_file('exchange.yaml', '\n'.join([
            'compartment: e',
            'compounds:',
            '  - id: cpd_A',
            '    lower: -42'
        ]))

        path = os.path.join(self._model_dir, 'fake.yaml')
        exchange = list(native.parse_exchange_list(path, [
            {'include': 'exchange.yaml'},
            {
                'compartment': 'e',
                'compounds': [
                    {'id': 'cpd_B', 'upper': 767}
                ]
            }
        ], 'e'))

        self.assertEqual(exchange, [
            (Compound('cpd_A', 'e'), None, -42, None),
            (Compound('cpd_B', 'e'), None, None, 767)
        ])

    def test_parse_limits_table_file(self):
        path = self.write_model_file('limits_1.tsv', '\n'.join([
            '# comment',
            '',
            'rxn_1',
            'rxn_2\t-100',
            'rxn_3\t-\t3e-1',
            'rxn_4\t-1000\t-100  # line comment'
        ]))

        limits = list(native.parse_limits_file(path))
        self.assertEqual(limits[0], ('rxn_1', None, None))
        self.assertEqual(limits[1], ('rxn_2', -100, None))
        self.assertEqual(limits[2][0], 'rxn_3')
        self.assertEqual(limits[2][1], None)
        self.assertAlmostEqual(limits[2][2], 3e-1)
        self.assertEqual(limits[3], ('rxn_4', -1000, -100))

    def test_parse_limits_table_file_too_many_fields(self):
        path = self.write_model_file('limits.tsv', '\n'.join([
            'rxn_1\t-\t100\ttext']))
        with self.assertRaises(native.ParseError):
            list(native.parse_limits_file(path))

    def test_parse_limits_yaml_file(self):
        path = self.write_model_file('limits_1.yaml', '\n'.join([
            '- include: limits_2.yml',
            '- reaction: rxn_3',
            '  lower: -1000',
            '- reaction: rxn_4',
            '  fixed: 10'
        ]))

        self.write_model_file('limits_2.yml', '\n'.join([
            '- reaction: rxn_1',
            '  lower: -1',
            '  upper: 25.5',
            '- reaction: rxn_2'
        ]))

        limits = list(native.parse_limits_file(path))
        self.assertEqual(limits, [
            ('rxn_1', -1, 25.5),
            ('rxn_2', None, None),
            ('rxn_3', -1000, None),
            ('rxn_4', 10, 10)
        ])

    def test_parse_model_table_file(self):
        path = self.write_model_file('model_1.tsv', '\n'.join([
            '# comment',
            'rxn_1',
            'rxn_2',
            'rxn_3',
            'rxn_4  # line comment']))

        reactions = list(native.parse_model_file(path))
        self.assertEqual(reactions, ['rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'])

    def test_parse_model_yaml_file(self):
        path = self.write_model_file('model_1.yaml', '''---
            - include: model_2.yaml
            - reactions:
              - rxn_3
              - rxn_4''')

        self.write_model_file('model_2.yaml', '''---
            - groups:
              - name: First group
                reactions: [rxn_1]
              - name: Second group
                reactions: [rxn_2]''')

        reactions = list(native.parse_model_file(path))
        self.assertEqual(reactions, ['rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'])


class TestCheckId(unittest.TestCase):
    def test_check_id_none(self):
        with self.assertRaises(native.ParseError):
            native._check_id(None, 'Compound')

    def test_check_id_not_string(self):
        with self.assertRaises(native.ParseError):
            native._check_id(False, 'Compound')

    def test_check_id_empty_string(self):
        with self.assertRaises(native.ParseError):
            native._check_id('', 'Reaction')

    def test_check_id_success(self):
        native._check_id(u'\u222b', 'Compound')


class TestNativeModel(unittest.TestCase):
    def test_properties(self):
        model = native.NativeModel()
        model.name = 'Test model'
        model.version_string = '1.0'
        model.biomass_reaction = 'rxn_1'
        model.extracellular_compartment = 'e'
        model.default_compartment = 'c'
        model.default_flux_limit = 1000
        self.assertEqual(model.name, 'Test model')
        self.assertEqual(model.version_string, '1.0')
        self.assertEqual(model.biomass_reaction, 'rxn_1')
        self.assertEqual(model.extracellular_compartment, 'e')
        self.assertEqual(model.default_compartment, 'c')
        self.assertEqual(model.default_flux_limit, 1000)


class TestNativeModelWriter(unittest.TestCase):
    def setUp(self):
        self.writer = native.ModelWriter()

    def test_convert_compartment_entry(self):
        compartment = entry.DictCompartmentEntry({
            'id': 'cytosol',
            'name': 'Cytosol',
            'custom': 456
        })
        d = self.writer.convert_compartment_entry(compartment, 'periplasm')
        self.assertIsInstance(d, OrderedDict)
        self.assertEqual(d['id'], 'cytosol')
        self.assertEqual(d['name'], 'Cytosol')
        self.assertEqual(d['custom'], 456)
        self.assertEqual(d['adjacent_to'], 'periplasm')

    def test_convert_compound_entry(self):
        compound = entry.DictCompoundEntry({
            'id': 'c1',
            'name': 'Compound 1',
            'charge': 2,
            'formula': None,
            'custom': 'ABC'
        })
        d = self.writer.convert_compound_entry(compound)
        self.assertIsInstance(d, OrderedDict)
        self.assertEqual(d['id'], 'c1')
        self.assertEqual(d['name'], 'Compound 1')
        self.assertEqual(d['charge'], 2)
        self.assertEqual(d['custom'], 'ABC')
        self.assertNotIn('formula', d)

    def test_convert_reaction_entry(self):
        equation = Reaction(
            Direction.Both, {Compound('c1', 'c'): -1, Compound('c2', 'c'): 1})
        reaction = entry.DictReactionEntry({
            'id': 'rxn01234',
            'name': 'Test reaction',
            'equation': equation,
            'custom_property': -34,
            'another_custom_property': None
        })
        d = self.writer.convert_reaction_entry(reaction)
        self.assertIsInstance(d, OrderedDict)
        self.assertEqual(d['id'], 'rxn01234')
        self.assertEqual(d['name'], 'Test reaction')
        self.assertEqual(d['equation'], equation)
        self.assertEqual(d['custom_property'], -34)
        self.assertNotIn('another_custom_property', d)

    def test_write_compartments(self):
        stream = StringIO()
        compartments = [
            entry.DictCompartmentEntry({
                'id': 'cytosol',
                'name': 'Cytosol',
                'custom': 456
            }),
            entry.DictCompartmentEntry({
                'id': 'periplasm',
                'name': 'Periplasm',
                'test': 'abc'
            })
        ]
        adjacencies = {
            'cytosol': 'periplasm',
            'periplasm': ['cytosol', 'e']
        }
        self.writer.write_compartments(stream, compartments, adjacencies)

        self.assertEqual(yaml.safe_load(stream.getvalue()), [
            {
                'id': 'cytosol',
                'adjacent_to': 'periplasm',
                'name': 'Cytosol',
                'custom': 456
            },
            {
                'id': 'periplasm',
                'adjacent_to': ['cytosol', 'e'],
                'name': 'Periplasm',
                'test': 'abc'
            }
        ])

    def test_write_compounds(self):
        stream = StringIO()
        compounds = [
            entry.DictCompoundEntry({
                'id': 'c1',
                'name': 'Compound 1',
                'charge': 2,
                'custom': 34.5
            }),
            entry.DictCompoundEntry({
                'id': 'c2',
                'name': 'Compound 2',
                'formula': Formula.parse('H2O')
            })
        ]
        self.writer.write_compounds(stream, compounds)

        self.assertEqual(yaml.safe_load(stream.getvalue()), [
            {
                'id': 'c1',
                'name': 'Compound 1',
                'charge': 2,
                'custom': 34.5
            },
            {
                'id': 'c2',
                'name': 'Compound 2',
                'formula': 'H2O'
            }
        ])

    def test_write_compounds_with_properties(self):
        stream = StringIO()
        compounds = [
            entry.DictCompoundEntry({
                'id': 'c1',
                'name': 'Compound 1',
                'charge': 2
            }),
            entry.DictCompoundEntry({
                'id': 'c2',
                'name': 'Compound 2',
                'formula': 'H2O'
            })
        ]
        self.writer.write_compounds(stream, compounds, properties={'name'})

        self.assertEqual(yaml.safe_load(stream.getvalue()), [
            {
                'id': 'c1',
                'name': 'Compound 1'
            },
            {
                'id': 'c2',
                'name': 'Compound 2'
            }
        ])

    def test_write_reactions(self):
        stream = StringIO()
        reactions = [
            entry.DictReactionEntry({
                'id': 'r1',
                'name': 'Reaction 1',
                'equation': Reaction(Direction.Both, {})
            }),
            entry.DictReactionEntry({
                'id': 'r2',
                'name': 'Reaction 2',
                'equation': Reaction(Direction.Forward, {
                    Compound('c1', 'c'): -1,
                    Compound('c2', 'c'): 2
                })
            })
        ]
        self.writer.write_reactions(stream, reactions)

        # The reaction equation for r1 is invalid (no compounds) and is
        # therefore skipped.
        self.assertEqual(yaml.safe_load(stream.getvalue()), [
            {
                'id': 'r1',
                'name': 'Reaction 1'
            },
            {
                'id': 'r2',
                'name': 'Reaction 2',
                'equation': 'c1[c] => (2) c2[c]'
            }
        ])


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2020  Elysha Sameth <esameth1@my.uri.edu>

from __future__ import unicode_literals

import unittest

from psamm import fluxanalysis
from psamm.lpsolver import generic
from psamm.datasource import native
from psamm.database import DictDatabase
from psamm.metabolicmodel import MetabolicModel
from psamm.datasource.reaction import parse_reaction
from psamm.commands.robustness import RobustnessTaskHandler, \
    RobustnessTaskHandlerFva

from six import itervalues


class TestRobustnessTaskHandler(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        self._problem = fluxanalysis.FluxBalanceProblem(
            self.model, self.solver)
        self._reactions = list(self.model.reactions)

    def test1_loop_removal_none(self):
        fluxes = RobustnessTaskHandler.handle_task(RobustnessTaskHandler(
            self.model, self.solver, 'none', self._reactions),
            ['rxn_6', 1000], 'rxn_6')

        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test2_loop_removal_l1min(self):
        fluxes = RobustnessTaskHandler.handle_task(RobustnessTaskHandler(
            self.model, self.solver, 'l1min', self._reactions),
            ['rxn_6', 1000], 'rxn_6')

        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'], 1000)
        self.assertAlmostEqual(fluxes['rxn_4'], 0)
        self.assertAlmostEqual(fluxes['rxn_5'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test3_loop_removal_tfba(self):
        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MIQP solver for tests')

        fluxes = RobustnessTaskHandler.handle_task(RobustnessTaskHandler(
            self.model, self.solver, 'tfba', self._reactions),
            ['rxn_3', 100], 'rxn_1')

        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'], 100)
        self.assertAlmostEqual(fluxes['rxn_4'], 900)
        self.assertAlmostEqual(fluxes['rxn_5'], 900)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test4_no_reactions(self):
        self._reactions2 = None
        fluxes = RobustnessTaskHandler.handle_task(RobustnessTaskHandler(
            self.model, self.solver, 'none', self._reactions2),
            ['rxn_6', 10], 'rxn_6')
        self.assertEqual(fluxes, 10)


class TestRobustnessTaskHandlerFva(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))

        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        self.model.limits['rxn_5'].upper = 100

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        self._problem = fluxanalysis.FluxBalanceProblem(
            self.model, self.solver)
        self._reactions = list(self.model.reactions)

    def test1_loop_removal_none(self):
        fluxes = RobustnessTaskHandlerFva.handle_task(RobustnessTaskHandlerFva(
            self.model, self.solver, 'none', self._reactions),
            ['rxn_6', 200], 'rxn_6')

        for bounds in itervalues(fluxes):
            self.assertEqual(len(bounds), 2)

        self.assertAlmostEqual(fluxes['rxn_1'][0], 100)

        self.assertAlmostEqual(fluxes['rxn_2'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_2'][1], 0)

        self.assertAlmostEqual(fluxes['rxn_5'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_5'][1], 100)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)

        self.assertGreater(fluxes['rxn_7'][1], 0)
        self.assertGreater(fluxes['rxn_8'][1], 0)

    def test2_loop_removal_tfba(self):
        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MIQP solver for tests')

        fluxes = RobustnessTaskHandlerFva.handle_task(RobustnessTaskHandlerFva(
            self.model, self.solver, 'tfba', self._reactions),
            ['rxn_6', 200], 'rxn_3')

        for bounds in itervalues(fluxes):
            self.assertEqual(len(bounds), 2)

        self.assertAlmostEqual(fluxes['rxn_1'][0], 100)

        self.assertAlmostEqual(fluxes['rxn_2'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_2'][1], 0)

        self.assertAlmostEqual(fluxes['rxn_3'][0], 200)

        self.assertAlmostEqual(fluxes['rxn_5'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_5'][1], 0)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)

        self.assertAlmostEqual(fluxes['rxn_7'][1], 0)
        self.assertAlmostEqual(fluxes['rxn_8'][1], 0)

    def test3_no_reactions(self):
        self._reactions2 = None
        fluxes = RobustnessTaskHandlerFva.handle_task(RobustnessTaskHandlerFva(
            self.model, self.solver, 'none', self._reactions2),
            ['rxn_6', 200], 'rxn_6')

        for bounds in itervalues(fluxes):
            self.assertEqual(len(bounds), 2)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)
        self.assertAlmostEqual(fluxes['rxn_6'][1], 200)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Base objects for representation of LP problems.

A linear programming problem is built from a number of constraints and an
objective function. The objective function is a linear expression represented
by :class:`.Expression`. The constraints are represented by :class:`.Relation`,
created from a linear expression and a relation sense (equals, greater, less).

Expressions are built from variables defined in the :class:`.Problem` instance.
In addition, an expression can contain a :class:`.VariableSet` instead of a
single variable. This allows many similar expressions to be represented by one
:class:`.Expression` instance which means that the LP problem can be
constructed faster.
"""

from __future__ import unicode_literals

import math
import numbers
import operator
from collections import Counter, defaultdict
import abc
import enum
from fractions import Fraction
from decimal import Decimal

import six
from six import add_metaclass, iteritems, viewkeys, viewitems, text_type
from six.moves import range, reduce

_INF = float('inf')


class SolverError(Exception):
    """Error wrapping solver specific errors."""


class VariableSet(tuple):
    """A tuple used to represent sets of variables."""


class Product(tuple):
    """A tuple used to represent a variable product."""


class _RangedAccessor(object):
    """Accessor to value bounded by minimum/maximum."""
    def __init__(self, obj, prop):
        self._obj = obj
        self._prop = prop

    @property
    def value(self):
        """Value of property."""
        if self._prop.fget is None:
            raise AttributeError('Unable to read attribute')
        return self._prop.fget(self._obj)

    @value.setter
    def value(self, v):
        if self._prop.fset is None:
            raise AttributeError('Unable to write attribute')
        self._prop.fset(self._obj, v)

    @value.deleter
    def value(self):
        if self._prop.fdel is None:
            raise AttributeError('Unable to delete attribute')
        self._prop.fdel(self._obj)

    @property
    def min(self):
        """Minimum value."""
        if self._prop.fmin is None:
            return -_INF
        return self._prop.fmin(self._obj)

    @property
    def max(self):
        """Maximum value."""
        if self._prop.fmax is None:
            return _INF
        return self._prop.fmax(self._obj)


class RangedProperty(object):
    """Numeric property with minimum and maximum values.

    The value attribute is used to get/set the actual value of the propery.
    The min/max attributes are used to get the bounds. The range is not
    automatically enforced when the value is set.
    """
    def __init__(self, fget=None, fset=None, fdel=None,
                 fmin=None, fmax=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        self.fmin = fmin
        self.fmax = fmax

        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return _RangedAccessor(obj, self)

    def __set__(self, obj, value):
        raise AttributeError('Unable to set attribute (try value property)')

    def __delete__(self, obj):
        raise AttributeError('Unable to delete attribute (try value property)')

    def getter(self, fget):
        return type(self)(
            fget, self.fset, self.fdel, self.fmin, self.fmax, self.__doc__)

    def setter(self, fset):
        return type(self)(
            self.fget, fset, self.fdel, self.fmin, self.fmax, self.__doc__)

    def deleter(self, fdel):
        return type(self)(
            self.fget, self.fset, fdel, self.fmin, self.fmax, self.__doc__)


def ranged_property(min=None, max=None):
    """Decorator for creating ranged property with fixed bounds."""
    min_value = -_INF if min is None else min
    max_value = _INF if max is None else max
    return lambda fget: RangedProperty(
        fget, fmin=lambda obj: min_value, fmax=lambda obj: max_value)


@six.python_2_unicode_compatible
class Expression(object):
    """Represents a linear expression

    The variables can be any hashable objects. If one or more variables are
    instead :class:`VariableSets <.VariableSet>`, this will be taken to
    represent a set of expressions separately using a different element of the
    :class:`.VariableSet`.

    >>> e = Expression({'x': 2, 'y': 3})
    >>> str(e)
    '2*x + 3*y'

    In order to provide a more natural syntax for creating
    :class:`Relations <.Relation>` the binary relation operators have been
    overloaded to return :class:`.Relation` instances.

    >>> rel = Expression({'x': 2}) >= Expression({'y': 3})
    >>> str(rel)
    '2*x - 3*y >= 0'

    .. warning::

        Chained relations cannot be converted to multiple
        relations, e.g. ``4 <= e <= 10`` will fail to produce the intended
        relations!
    """

    def __init__(self, variables={}, offset=0):
        self._variables = Counter(variables)
        self._offset = offset

    @property
    def offset(self):
        """Value of the offset"""
        return self._offset

    def variables(self):
        """Return immutable view of variables in expression."""
        return viewkeys(self._variables)

    def values(self):
        """Return immutable view of (variable, value)-pairs in expression."""
        return viewitems(self._variables)

    def value(self, variable):
        return self._variables.get(variable, 0)

    def value_sets(self):
        """Iterator of expression sets

        This will yield an iterator of (variable, value)-pairs for each
        expression in the expression set (each equivalent to values()). If none
        of the variables is a set variable then a single iterator will be
        yielded.
        """
        count = max(1 if not isinstance(var, VariableSet) else
                    len(var) for var in self._variables)

        def value_set(n):
            for variable, value in iteritems(self._variables):
                if isinstance(variable, VariableSet):
                    yield variable[n], value
                else:
                    yield variable, value
        for i in range(count):
            yield value_set(i)

    def __contains__(self, variable):
        return variable in self._variables

    def __add__(self, other):
        """Add expression with a number or another expression"""

        if isinstance(other, numbers.Number):
            if math.isinf(self._offset) or math.isinf(other):
                return self.__class__(offset=self._offset + other)
            return self.__class__(self._variables, self._offset + other)
        elif isinstance(other, self.__class__):
            if math.isinf(self._offset) or math.isinf(other._offset):
                return self.__class__(offset=self._offset + other._offset)
            variables = Counter(self._variables)
            variables.update(other._variables)
            return self.__class__(variables, self._offset + other._offset)
        return NotImplemented

    __radd__ = __add__

    def __iadd__(self, other):
        if isinstance(other, numbers.Number):
            if math.isinf(self._offset) or math.isinf(other):
                self._variables = {}
            self._offset += other
        elif isinstance(other, self.__class__):
            self._offset += other._offset
            if math.isinf(self._offset) or math.isinf(other._offset):
                self._variables = {}
            else:
                self._variables.update(other._variables)
        else:
            return NotImplemented

        return self

    def __sub__(self, other):
        return self + -other

    def __rsub__(self, other):
        return -self + other

    def __isub__(self, other):
        self += -other
        return self

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            if math.isinf(other):
                return self.__class__(offset=float('nan'))

            return self.__class__(
                {var: value * other for var, value in
                 iteritems(self._variables)}, self._offset * other)
        elif isinstance(other, self.__class__):
            variables = defaultdict(int)
            for v1, value1 in iteritems(self._variables):
                for v2, value2 in iteritems(other._variables):
                    p1 = v1 if isinstance(v1, Product) else Product((v1,))
                    p2 = v2 if isinstance(v2, Product) else Product((v2,))
                    product = Product(sorted(p1 + p2, key=text_type))
                    variables[product] += value1 * value2

                if other._offset != 0:
                    variables[v1] += value1 * other._offset

            if self._offset != 0:
                for v2, value2 in iteritems(other._variables):
                    variables[v2] += value2 * self._offset

            offset = self._offset * other._offset
            return self.__class__(variables, offset=offset)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __imul__(self, other):
        if isinstance(other, numbers.Number):
            if math.isinf(other):
                self._variables = {}
                self._offset = float('nan')
            else:
                for var in self._variables:
                    self._variables[var] *= other
            self._offset *= other
        elif isinstance(other, self.__class__):
            variables = defaultdict(int)
            for v1, value1 in iteritems(self._variables):
                for v2, value2 in iteritems(other._variables):
                    p1 = v1 if isinstance(v1, Product) else Product((v1,))
                    p2 = v2 if isinstance(v2, Product) else Product((v2,))
                    product = Product(sorted(p1 + p2, key=text_type))
                    variables[product] += value1 * value2

                if other._offset != 0:
                    variables[v1] += value1 * other._offset

            if self._offset != 0:
                for v2, value2 in iteritems(other._variables):
                    variables[v2] += value2 * self._offset

            self._offset *= other._offset
            self._variables = dict(variables)
        else:
            return NotImplemented

        return self

    def __neg__(self):
        return self.__class__(
            {var: -value for var, value in iteritems(self._variables)},
            -self._offset)

    def __pow__(self, other):
        if isinstance(other, int):
            if other < 0:
                raise ValueError('Exponent must be positive')
            elif other == 0:
                return self.__class__(offset=1)

            r = self.__class__(self._variables, self._offset)
            for i in range(other - 1):
                r.__imul__(self)

            return r
        else:
            return NotImplemented

    def __rpow__(self, other):
        return pow(other, self)

    def __ipow__(self, other):
        if isinstance(other, int):
            if other < 0:
                raise ValueError('Exponent must be positive')
            elif other == 0:
                self._variables = {}
                self._offset = 1
            else:
                tmp = self.__class__(self._variables, self._offset)
                for i in range(other - 1):
                    self.__imul__(tmp)
        else:
            return NotImplemented

        return self

    def __eq__(self, other):
        """Return equality relation (equation): self == other

        This method is overloaded so that relations can be
        formed using a natural syntax.
        """

        return Relation(RelationSense.Equals, self - other)

    def __ge__(self, other):
        """Return greater-than relation (inequality): self >= other

        This method is overloaded so that relations can be
        formed using a natural syntax.
        """

        return Relation(RelationSense.Greater, self - other)

    def __le__(self, other):
        """Return less-than relation (inequality): self <= other

        This method is overloaded so that relations can be
        formed using a natural syntax.
        """

        return Relation(RelationSense.Less, self - other)

    def __gt__(self, other):
        """Return strictly greater-than relation (inequality): self > other

        This method is overloaded so that relations can be
        formed using a natural syntax.
        """

        return Relation(RelationSense.StrictlyGreater, self - other)

    def __lt__(self, other):
        """Return strictly less-than relation (inequality): self < other

        This method is overloaded so that relations can be
        formed using a natural syntax.
        """

        return Relation(RelationSense.StrictlyLess, self - other)

    def __str__(self):
        """Return string representation of expression"""

        def sort_key_variable(item):
            key, _ = item
            is_product = isinstance(key, Product)
            return is_product, text_type(key)

        def all_terms():
            count_vars = 0
            for name, value in sorted(
                    iteritems(self._variables), key=sort_key_variable):
                if value != 0:
                    count_vars += 1
                    if isinstance(name, VariableSet):
                        yield '<set>', value
                    elif isinstance(name, Product):
                        yield '*'.join(text_type(n) for n in name), value
                    else:
                        yield name, value
            if self._offset != 0 or count_vars == 0:
                yield None, self._offset

        terms = []
        for i, (name, value) in enumerate(all_terms()):
            if i == 0:
                # First term is special
                if name is None:
                    terms.append('{}'.format(value))
                elif abs(value) == 1:
                    terms.append(
                        text_type(name) if value > 0 else '-{}'.format(name))
                else:
                    terms.append('{}*{}'.format(value, name))
            else:
                prefix = '+' if value >= 0 else '-'
                if name is None:
                    terms.append('{} {}'.format(prefix, abs(value)))
                elif abs(value) == 1:
                    terms.append('{} {}'.format(prefix, name))
                else:
                    terms.append('{} {}*{}'.format(prefix, abs(value), name))
        return ' '.join(terms)

    def __repr__(self):
        return str('<{} {}>').format(
            self.__class__.__name__, repr(str(self)))


@enum.unique
class RelationSense(enum.Enum):
    Equals = '=='
    Greater = '>='
    Less = '<='
    StrictlyGreater = '>'
    StrictlyLess = '<'


@six.python_2_unicode_compatible
class Relation(object):
    """Represents a binary relation (equation or inequality)

    Relations can be equalities or inequalities. All relations
    of this type can be represented as a left-hand side expression
    and the type of relation. In this representation, the right-hand
    side is always zero.
    """

    def __init__(self, sense, expression):
        self._sense = sense
        self._expression = expression

    @property
    def sense(self):
        """Type of relation (equality or inequality)

        Can be one of Equal, Greater or Less, or one of the
        strict relations, StrictlyGreater or StrictlyLess.
        """
        return self._sense

    @property
    def expression(self):
        """Left-hand side expression"""
        return self._expression

    def __str__(self):
        """Convert relation to string representation"""
        var_expr = Expression(dict(self._expression.values()))
        return '{} {} {}'.format(
            str(var_expr), self._sense.value, -self._expression.offset)

    def __repr__(self):
        return str('<{} {}>').format(self.__class__.__name__, repr(str(self)))

    def __nonzero__(self):
        # Override __nonzero__ (and __bool__) here so we can avoid bugs when a
        # user mistakenly thinks that "4 <= x <= 100" should work.
        # This kind of expression expands to "4 <= x and x <= 100" and since we
        # cannot override the "and" operator, the relation "4 <= x" is
        # converted to a bool. Override here to raise an error to make sure
        # that this syntax always fails.
        raise ValueError('Unable to convert relation to bool')

    __bool__ = __nonzero__


@enum.unique
class ObjectiveSense(enum.Enum):
    """Enumeration of objective sense values"""

    Minimize = -1
    """Minimize objective function"""

    Maximize = 1
    """Maximize objective function"""


@enum.unique
class VariableType(enum.Enum):
    """Enumeration of variable types"""

    Continuous = 'C'
    """Continuous variable type"""

    Integer = 'I'
    """Integer variable type"""

    Binary = 'B'
    """Binary variable type (0 or 1)"""


@add_metaclass(abc.ABCMeta)
class Solver(object):
    """Factory for LP Problem instances"""

    @abc.abstractmethod
    def create_problem(self):
        """Create a new :class:`.Problem` instance"""


@add_metaclass(abc.ABCMeta)
class Constraint(object):
    """Represents a constraint within an LP Problem"""

    @abc.abstractmethod
    def delete(self):
        """Remove constraint from Problem instance"""


class VariableNamespace(object):
    """Namespace for defining variables.

    Namespaces are always unique even if two namespaces have the same name.
    Variables defined within a namespace will never clash with a global
    variable name or a name defined in another namespace. Namespaces should
    be created using the :meth:`.Problem.namespace`.

    >>> v = prob.namespace(name='v')
    >>> v.define([1, 2, 5], lower=0, upper=10)
    >>> prob.set_objective(v[1] + 3*v[2] - 5 * v[5])
    """
    def __init__(self, problem, **kwargs):
        self._name = kwargs.pop('name', None)
        self._problem = problem
        self._define_kwargs = kwargs

    def define(self, names, **kwargs):
        """Define variables within the namespace.

        This is similar to :meth:`.Problem.define` except that names must be
        given as an iterable. This method accepts the same keyword arguments
        as :meth:`.Problem.define`.
        """
        define_kwargs = dict(self._define_kwargs)
        define_kwargs.update(kwargs)
        self._problem.define(
            *((self, name) for name in names), **define_kwargs)

    def __getitem__(self, key):
        return self._problem.var((self, key))

    def __contains__(self, key):
        return self._problem.has_variable((self, key))

    def has_variable(self, name):
        return self._problem.has_variable((self, name))

    def __call__(self, name):
        return self._problem.var((self, name))

    def set(self, names):
        """Return a variable set of the given names in the namespace.

        >>> v = prob.namespace(name='v')
        >>> v.define([1, 2, 5], lower=0, upper=10)
        >>> prob.add_linear_constraints(v.set([1, 2]) >= 4)
        """
        return self._problem.set((self, name) for name in names)

    def sum(self, names):
        """Return the sum of the given names in the namespace.

        >>> v = prob.namespace(name='v')
        >>> v.define([1, 2, 5], lower=0, upper=10)
        >>> prob.set_objective(v.sum([2, 5]))  # v[2] + v[5]
        """
        return Expression({(self, name): 1 for name in names})

    def expr(self, items):
        """Return the sum of each name multiplied by a coefficient.

        >>> v = prob.namespace(name='v')
        >>> v.define(['a', 'b', 'c'], lower=0, upper=10)
        >>> prob.set_objective(v.expr([('a', 2), ('b', 1)]))
        """
        return Expression({(self, name): value for name, value in items})

    def value(self, name):
        """Return value of given variable in namespace.

        >>> v = prob.namespace(name='v')
        >>> v.define([1, 2, 5], lower=0, upper=10)
        >>> prob.solve()
        >>> print(v.value(2))
        """
        return self._problem.result.get_value((self, name))

    def __repr__(self):
        name = repr(self._name) if self._name is not None else 'Unnamed'
        return str('<{}: {}>').format(self.__class__.__name__, name)


@add_metaclass(abc.ABCMeta)
class Problem(object):
    """Representation of LP Problem instance

    Variable names in the problem can be any hashable object. It is the
    responsibility of the solver interface to translate the object into a
    unique string if required by the underlying LP solver.
    """

    @abc.abstractmethod
    def define(self, *names, **kwargs):
        """Define a variable in the problem.

        Variables must be defined before they can be accessed by var() or
        set(). This function takes keyword arguments lower and upper to define
        the bounds of the variable (default: -inf to inf). The keyword argument
        types can be used to select the type of the variable (Continuous
        (default), Binary or Integer). Setting any variables different than
        Continuous will turn the problem into an MILP problem. Raises
        ValueError if a name is already defined.
        """

    @abc.abstractmethod
    def has_variable(self, name):
        """Check whether a variable is defined in the problem."""

    def namespace(self, names=None, **kwargs):
        """Return namespace for this problem.

        If names is given it should be an iterable of names to define in the
        namespace. Other keyword arguments can be specified which will be
        used to define the names given as well as being used as default
        parameters for names that are defined later.

        >>> v = prob.namespace(name='v')
        >>> v.define([1, 2, 5], lower=0, upper=10)
        >>> prob.set_objective(v[1] + 3*v[2] - 5 * v[5])
        """
        ns = VariableNamespace(self, **kwargs)
        if names is not None:
            ns.define(names)
        return ns

    def var(self, name):
        """Return variable as an :class:`.Expression`."""
        if not self.has_variable(name):
            raise ValueError('Undefined variable: {}'.format(name))
        return Expression({name: 1})

    def expr(self, values, offset=0):
        """Return the given dictionary of values as an :class:`.Expression`."""
        if isinstance(values, dict):
            for name in values:
                if not self.has_variable(name):
                    raise ValueError('Undefined variable: {}'.format(name))
            return Expression(values, offset=offset)

        if not self.has_variable(values):
            raise ValueError('Undefined variable: {}'.format(values))
        return Expression({values: 1}, offset=offset)

    def set(self, names):
        """Return variables as a set expression.

        This returns an :class:`.Expression` containing a
        :class:`.VariableSet`.
        """
        names = tuple(names)
        if any(not self.has_variable(name) for name in names):
            raise ValueError('Undefined variables: {}'.format(
                set(names) - set(self._variables)))
        return Expression({VariableSet(names): 1})

    def _check_relation(self, relation):
        """Check whether the given relation is valid.

        Return false normally, or true if the relation is determined to be
        tautologically true. Raises ``ValueError`` if the relation is
        tautologically false. This includes relations given as a ``False`` bool
        value or certain types of relations that have an infinity offset.
        """
        if isinstance(relation, bool):
            if not relation:
                raise ValueError('Unsatisfiable relation added')
            return True

        if relation.sense in (
                RelationSense.StrictlyGreater, RelationSense.StrictlyLess):
            raise ValueError(
                'Strict relations are invalid in LP-problems:'
                ' {}'.format(relation))

        if relation.expression.offset == -_INF:
            if relation.sense == RelationSense.Less:
                return True
            else:
                raise ValueError('Unsatisfiable relation added: {}'.format(
                    relation))
        elif relation.expression.offset == _INF:
            if relation.sense == RelationSense.Greater:
                return True
            else:
                raise ValueError('Unsatisfiable relation added: {}'.format(
                    relation))

        return False

    @abc.abstractmethod
    def add_linear_constraints(self, *relations):
        """Add constraints to the problem.

        Each constraint is given as a :class:`.Relation`, and the expression
        in that relation can be a set expression. Returns a sequence of
        :class:`Constraints <.Constraint>`.
        """

    @abc.abstractmethod
    def set_objective(self, expression):
        """Set objective of the problem to the given :class:`.Expression`."""

    set_linear_objective = set_objective
    """Set objective of the problem.

    .. deprecated:: 0.19
       Use :meth:`set_objective` instead.
    """

    @abc.abstractmethod
    def set_objective_sense(self, sense):
        """Set type of problem (minimize or maximize)"""

    def solve(self, sense=None):
        """Solve problem and return result.

        Raises :class:`SolverError` if the solution is not optimal.
        """
        result = self.solve_unchecked(sense)
        if not result:
            raise SolverError('Solution not optimal: {}'.format(result.status))

        return result

    @abc.abstractmethod
    def solve_unchecked(self, sense=None):
        """Solve problem and return result.

        The user must manually check the status of the result to determine
        whether an optimal solution was found. A :class:`SolverError` may still
        be raised if the underlying solver raises an exception.
        """

    @abc.abstractproperty
    def result(self):
        """Result of solved problem"""

    @ranged_property(min=None, max=None)
    def feasibility_tolerance(self):
        raise NotImplementedError('Feasiblity tolerance not available')

    @ranged_property(min=None, max=None)
    def optimality_tolerance(self):
        raise NotImplementedError('Optimality tolerance not available')

    @ranged_property(min=None, max=None)
    def integrality_tolerance(self):
        raise NotImplementedError('Integrality tolerance not available')


class InvalidResultError(Exception):
    """Raised when a result that has been invalidated is accessed"""

    def __init__(self, msg=None):
        if msg is None:
            msg = ('Previous result is no longer valid as problem'
                   ' has been solved again')
        super(InvalidResultError, self).__init__(msg)


class Result(object):
    """Result of solving an LP problem

    The result is tied to the solver instance and is valid at least until the
    problem is solved again. If the problem has been solved again an
    :class:`InvalidResultError` may be raised.
    """

    @abc.abstractproperty
    def success(self):
        """Whether solution was optimal"""

    def __nonzero__(self):
        """Whether solution was optimal"""
        return self.success

    __bool__ = __nonzero__

    @abc.abstractproperty
    def status(self):
        """String indicating the status of the problem result"""

    @abc.abstractproperty
    def unbounded(self):
        """Whether solution is unbounded"""

    @abc.abstractmethod
    def _get_value(self, var):
        """Return the solution value of a single variable."""

    @abc.abstractmethod
    def _has_variable(self, var):
        """Return whether variable exists in the solution."""

    def _evaluate_expression(self, expr):
        """Evaluate an :class:`.Expression` using :meth:`_get_value`."""
        def cast_value(v):
            # Convert Decimal to Fraction to allow successful multiplication
            # by either float (most solvers) or Fraction (exact solver).
            # Multiplying Fraction and float results in a float, and
            # multiplying Fraction and Fraction result in Fraction, which are
            # exactly the types of results we want.
            if isinstance(v, Decimal):
                return Fraction(v)
            return v

        total = cast_value(expr.offset)
        for var, value in expr.values():
            value = cast_value(value)
            if not isinstance(var, Product):
                total += self._get_value(var) * value
            else:
                total += reduce(
                    operator.mul, (self._get_value(v) for v in var), value)
        return total

    def get_value(self, expression):
        """Get value of variable or expression in result

        Expression can be an object defined as a name in the problem, in which
        case the corresponding value is simply returned. If expression is an
        actual :class:`.Expression` object, it will be evaluated using the
        values from the result.
        """
        if isinstance(expression, Expression):
            return self._evaluate_expression(expression)
        elif not self._has_variable(expression):
            raise ValueError('Unknown expression: {}'.format(expression))

        return self._get_value(expression)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2020  Jing Wang <wjingsjtu@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest
import tempfile
import os
import shutil
from collections import namedtuple
import pandas as pd

from psamm import bayesian, bayesian_util as util
from psamm.datasource.native import (
    ModelReader, CompoundEntry, Formula, ReactionEntry)
from psamm.datasource.reaction import parse_reaction


class TestBayesianPredictor(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model1.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    name: rxn_1',
                '    equation: A[e] => B[c]',
                '    genes: gene1 and gene2 or gene3',
                '  - id: rxn_2',
                '    name: rxn_2',
                '    equation: B[c] => C[e]',
                '    genes: gene5 or gene6',
                '  - id: rxn_3',
                '    name: rxn_3',
                '    equation: A[e] + (6) B[c] <=> (6) C[e] + (6) D[c]',
                '    genes: gene7',
                'compounds:',
                '  - id: A',
                '    name: A',
                '    formula: C6H12O6',
                '    charge: 1',
                '  - id: B',
                '    name: B',
                '    formula: O2',
                '    charge: 1',
                '    kegg: C00010',
                '  - id: C',
                '    name: C',
                '    formula: CO2',
                '    charge: -1',
                '  - id: D',
                '    name: D',
                '    formula: H2O',
            ]))
        self._model1 = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'model1.yaml')).create_model()
        self._model1 = bayesian.MappingModel(self._model1)

        with open(os.path.join(self._model_dir, 'model2.yaml'), 'w') as f:
            f.write('\n'.join([
                '---',
                'reactions:',
                '  - id: rxn_1',
                '    name: rxn_1',
                '    equation: A[e] => B2[c]',
                '    genes: gene1 and gene2 or gene3',
                '  - id: rnx_2',
                '    equation: B2[c] => C[e]',
                '    genes: gene5 and gene6',
                '  - id: rxn_3',
                '    name: rxn_3',
                '    equation: A[e] <=> (6) C[e] + (6) D4[c]',
                '  - id: rxn_4',
                '    equation: E[c] => F[c]',
                'compounds:',
                '  - id: A',
                '    name: a',
                '    formula: C6H11O6',
                '    charge: 0',
                '  - id: B2',
                '    name: -B ()',
                '    formula: O2',
                '    charge: 1',
                '    kegg: C00010',
                '  - id: C',
                '    name: C',
                '    charge: -1',
                '  - id: D4',
                '    name: D',
                '    formula: H2O',
                '    charge: -1',
            ]))
        self._model2 = ModelReader.reader_from_path(
            os.path.join(self._model_dir, 'model2.yaml')).create_model()
        self._model2 = bayesian.MappingModel(self._model2)

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def test_summary(self):
        self._model1.print_summary()
        self._model2.print_summary()

    def test_consistency_check(self):
        self.assertTrue(self._model1.check_reaction_compounds())
        self.assertFalse(self._model2.check_reaction_compounds())

    def test_compound_map(self):
        p = bayesian.BayesianCompoundPredictor(
            self._model1, self._model2, 1, self._model_dir, True, True
        )
        self.assertEqual(p.model1, self._model1)
        self.assertEqual(p.model2, self._model2)
        self.assertEqual(p.map('A', 'A'), 1)
        self.assertIsInstance(p, bayesian.BayesianCompoundPredictor)
        best_map = p.get_best_map()
        self.assertIsInstance(best_map, pd.DataFrame)
        self.assertIsInstance(p.get_raw_map(), pd.DataFrame)
        self.assertIsInstance(p.get_cpd_pred(), pd.Series)
        self.assertLess(best_map.loc[('B', 'B2'), 'p_id'],
                        best_map.loc[('A', 'A'), 'p_id'])
        self.assertIn(('A', 'A'), best_map.index)
        self.assertIn(('B', 'B2'), best_map.index)
        self.assertIn(('C', 'C'), best_map.index)
        self.assertIn(('D', 'D4'), best_map.index)

    def test_reaction_map(self):
        cpd_pred = bayesian.BayesianCompoundPredictor(
            self._model1, self._model2, 1, self._model_dir, False, False
        ).get_cpd_pred()
        p = bayesian.BayesianReactionPredictor(
            self._model1, self._model2, cpd_pred,
            1, self._model_dir, True, True
        )
        self.assertIsInstance(p, bayesian.BayesianReactionPredictor)
        best_map = p.get_best_map()
        self.assertIsInstance(best_map, pd.DataFrame)
        self.assertIsInstance(p.get_raw_map(), pd.DataFrame)
        self.assertIn(('rxn_1', 'rxn_1'), best_map.index)
        self.assertIn(('rxn_2', 'rnx_2'), best_map.index)
        self.assertIn(('rxn_3', 'rxn_3'), best_map.index)

        p = bayesian.BayesianReactionPredictor(
            self._model1, self._model2, cpd_pred,
            1, self._model_dir, True, False
        )
        self.assertEqual(p.model1, self._model1)
        self.assertEqual(p.model2, self._model2)
        self.assertAlmostEqual(p.map('rxn_1', 'rxn_1'), 1, 1)
        self.assertIsInstance(p, bayesian.BayesianReactionPredictor)
        best_map = p.get_best_map()
        self.assertIsInstance(best_map, pd.DataFrame)
        self.assertIsInstance(p.get_raw_map(), pd.DataFrame)

    def test_compound_id_likelihood(self):
        match, unmatch = bayesian.compound_id_likelihood(
            CompoundEntry({'id': 'A'}), CompoundEntry({'id': 'A'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.compound_id_likelihood(
            CompoundEntry({'id': 'A'}), CompoundEntry({'id': 'B'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertLess(match, unmatch)

    def test_compound_name_likelihood(self):
        match, unmatch = bayesian.compound_name_likelihood(
            CompoundEntry({'id': 'A', 'name': 'Test'}),
            CompoundEntry({'id': 'B', 'name': 'te-s(t)'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.compound_name_likelihood(
            CompoundEntry({'id': 'A', 'name': 'Test'}),
            CompoundEntry({'id': 'B', 'name': 'Test1'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertLess(match, unmatch)

    def test_compound_charge_likelihood(self):
        match, unmatch = bayesian.compound_charge_likelihood(
            CompoundEntry({'id': 'A', 'charge': 1}),
            CompoundEntry({'id': 'B', 'charge': 1}),
            1.0 / 500, 0.5, 0.5
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.compound_charge_likelihood(
            CompoundEntry({'id': 'A', 'charge': 1}),
            CompoundEntry({'id': 'B', 'charge': -1}),
            1.0 / 500, 0.5, 0.5
        )
        self.assertLess(match, unmatch)
        match, unmatch = bayesian.compound_charge_likelihood(
            CompoundEntry({'id': 'A'}),
            CompoundEntry({'id': 'B', 'charge': -1}),
            1.0 / 500, 0.5, 0.5
        )
        self.assertEqual(match, unmatch)

    def test_compound_formula_likelihood(self):
        match, unmatch = bayesian.compound_formula_likelihood(
            CompoundEntry({'id': 'A', 'formula': Formula.parse('C3H4')}),
            CompoundEntry(
                {'id': 'B', 'formula': Formula.parse('C3H5'), 'charge': 1}),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.compound_formula_likelihood(
            CompoundEntry({'id': 'A', 'formula': Formula.parse('C3H4')}),
            CompoundEntry({'id': 'B', 'formula': Formula.parse('C3H5')}),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertLess(match, unmatch)
        match, unmatch = bayesian.compound_formula_likelihood(
            CompoundEntry({'id': 'A', 'formula': Formula.parse('C3H4')}),
            CompoundEntry({'id': 'B', 'formula': None}),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertEqual(match, unmatch)

    def test_compound_kegg_likelihood(self):
        CompoundEntry = namedtuple('CompoundEntry', ['id', 'kegg'])
        match, unmatch = bayesian.compound_kegg_likelihood(
            CompoundEntry(id='A', kegg='C00001'),
            CompoundEntry(id='B', kegg='C00001'),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.compound_kegg_likelihood(
            CompoundEntry(id='A', kegg='C00001'),
            CompoundEntry(id='B', kegg='C00003'),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertLess(match, unmatch)
        match, unmatch = bayesian.compound_kegg_likelihood(
            CompoundEntry(id='A', kegg=None),
            CompoundEntry(id='B', kegg='C00001'),
            1.0 / 500, 1.0 / 1000, 0.5
        )
        self.assertEqual(match, unmatch)

    def test_reaction_id_likelihood(self):
        match, unmatch = bayesian.reaction_id_likelihood(
            ReactionEntry({'id': 'r1'}),
            ReactionEntry({'id': 'r1'}),
            1.0 / 500, 1.0 / 500, 499.0 / 500
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.reaction_id_likelihood(
            ReactionEntry({'id': 'r1'}),
            ReactionEntry({'id': 'r2'}),
            1.0 / 500, 1.0 / 500, 499.0 / 500
        )
        self.assertLess(match, unmatch)

    def test_reaction_name_likelihood(self):
        match, unmatch = bayesian.reaction_name_likelihood(
            ReactionEntry({'id': 'r1', 'name': 'Test'}),
            ReactionEntry({'id': 'r1', 'name': 'te-s(t)'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.reaction_name_likelihood(
            ReactionEntry({'id': 'r1', 'name': 'Test'}),
            ReactionEntry({'id': 'r2', 'name': 'Test2'}),
            1.0 / 500, 1.0 / 1000
        )
        self.assertLess(match, unmatch)

    def test_reaction_genes_likelihood(self):
        match, unmatch = bayesian.reaction_genes_likelihood(
            ReactionEntry({'id': 'r1', 'genes': 'g1 and g2 or g3'}),
            ReactionEntry({'id': 'r1', 'genes': 'g2 and g1'}),
            1.0 / 500, 1.0 / 1000, 0.8
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.reaction_genes_likelihood(
            ReactionEntry({'id': 'r1', 'genes': 'g1 and g2 or g3'}),
            ReactionEntry({'id': 'r1', 'genes': 'g2 and g4'}),
            1.0 / 500, 1.0 / 1000, 0.8
        )
        self.assertLessEqual(match, unmatch)
        match, unmatch = bayesian.reaction_genes_likelihood(
            ReactionEntry({'id': 'r1', 'genes': 'g1 and g2 or g3'}),
            ReactionEntry({'id': 'r1', 'genes': 'g2 and g4'}),
            1.0 / 500, 1.0 / 1000, 0.8, {'g1': 'g4', 'g4': 'g1'}
        )
        self.assertGreater(match, unmatch)
        match, unmatch = bayesian.reaction_genes_likelihood(
            ReactionEntry({'id': 'r1', 'genes': None}),
            ReactionEntry({'id': 'r1', 'genes': 'g4 and g5'}),
            1.0 / 500, 1.0 / 1000, 0.8
        )
        self.assertEqual(match, unmatch)

    def test_reaction_equation_likelihood(self):
        cpd_map = {'A': set(['A1']),
                   'B': set(['B1']),
                   'C': set(['C1']),
                   'D': set(['D1'])}
        cpd_score = {'A': 1, 'B': 1, 'C': 0.9, 'D': 0.9}
        eq1 = parse_reaction('A[c] + C[c] + B[c] => D[c]')
        eq2 = parse_reaction('A1[c] + C1[c] + B1[c] => D1[c]')
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': eq2}),
            cpd_map, cpd_score
        )
        self.assertGreater(m, un)
        eq2 = parse_reaction('D1[c] + F1[c] <=> A1[c] + C1[c] + B1[c]')
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': eq2}),
            cpd_map, cpd_score
        )
        self.assertGreater(m, un)
        eq2 = parse_reaction('A1[c] => D1[c] + C1[c]')
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': eq2}),
            cpd_map, cpd_score
        )
        self.assertLess(m, un)
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': None}),
            cpd_map, cpd_score
        )
        self.assertEqual(m, un)
        eq2 = parse_reaction('A1[s] + C1[s] + B1[s] => D1[s]')
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': eq2}),
            cpd_map, cpd_score
        )
        self.assertLess(m, un)
        m, un = bayesian.reaction_equation_compound_mapping_likelihood(
            ReactionEntry({'id': 'r1', 'equation': eq1}),
            ReactionEntry({'id': 'r2', 'equation': eq2}),
            cpd_map, cpd_score,
            {'c': 's'}
        )
        self.assertGreater(m, un)


class Test_bayesian_util(unittest.TestCase):
    def test_name_equals(self):
        self.assertFalse(util.name_equals('Test', None))

    def test_name_similar(self):
        score = util.name_similar('kit-ten', 'sitt(i)ng')
        self.assertEqual(score, 1 - 3.0 / 7)
        score = util.name_similar('kitten', None)
        self.assertEqual(score, 0.01)
        score = util.name_similar('kitten', '')
        self.assertEqual(score, 0.01)

    def test_fomula_equals(self):
        self.assertFalse(util.formula_equals(None, None, 0, 1))

    def test_formula_exact(self):
        self.assertTrue(
            util.formula_exact(Formula.parse('C3H4O5'),
                               Formula.parse('C3H4O5'))
        )
        self.assertFalse(
            util.formula_exact(Formula.parse('C2H3O6'), None)
        )
        self.assertFalse(
            util.formula_exact(Formula.parse('C2H3O6'),
                               Formula.parse('C5H11'))
        )

    def test_pairwise_distance(self):
        for pair, score in util.pairwise_distance(['kitten'], ['sitting'],
                                                  util.jaccard):
            self.assertTupleEqual(
                (pair, score),
                (('kitten', 'sitting'), 3.0 / 7)
            )

    def test_gene_equals(self):
        g1 = '(A1 and B21) or (A12 and B2)'
        g2 = 'A12 and B2'
        g3 = 'A1 and B2'
        g4 = '(C and D) or (E and F)'
        gmap = {'A12': 'D', 'B2': 'C', 'C': 'B2', 'D': 'A12'}
        self.assertTrue(util.genes_equals(g1, g2))
        self.assertFalse(util.genes_equals(g1, g3))
        self.assertFalse(util.genes_equals(g1, None))
        self.assertTrue(util.genes_equals(g1, g4, gmap))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2019-2020 Jing Wang <jingwang89@uri.edu>

from __future__ import print_function
from collections import defaultdict
import argparse
import re
import csv
import logging
from os import path, mkdir

from ..command import Command
from ..expression import boolean
from psamm.datasource.native import ModelWriter
from psamm.importer import write_yaml_model
logger = logging.getLogger(__name__)


class PsammotateCommand(Command):
    """Generate draft model from a template using provided gene mapping.

    This function will take a provided reciprocal best hit file
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--rbh', metavar='file', type=argparse.FileType('r'),
            help=('A two column mapping of genes from the template '
                  'organism to the target organism.'))
        parser.add_argument(
            '--template', type=int,
            help=('The column of the RBH file where the template model '
                  'genes are listed, starts from 1.'))
        parser.add_argument(
            '--target', type=int,
            help=('The column of the RBH file where the target '
                  'model genes are listed, starts from 1.'))
        parser.add_argument(
            '--ignore-na', action='store_true',
            help=('Ignore the reactions that do not have gene association. '
                  '(default: retain these reactions in new reactions file)'))
        parser.add_argument(
            '--suffix', type=str, default=None,
            help='Suffix to append to end of reaction IDs and compartments.')
        g = parser.add_mutually_exclusive_group()
        g.add_argument(
            '--export-model', type=str, default=None,
            help='Path to directory for full model export. Cannot be specified'
                 ' with --output option.')
        g.add_argument(
            '--output', type=str, default='homolo_reactions',
            help=('The prefix of output YAML file, '
                  '(default: homolo_reactions). Cannot be specified with '
                  '--export-model option.'))
        super(PsammotateCommand, cls).init_parser(parser)

    def run(self):
        """Run psammotate command"""

        if self._args.export_model is None:
            if path.exists('{}.yaml'.format(self._args.output)):
                logger.warning('File {}.yaml already exists. '
                               'Please choose a different file name '
                               'through the --output '
                               'option.'.format(self._args.output))
                quit()
        elif self._args.export_model is not None:
            if path.exists('{}'.format(self._args.export_model)):
                logger.warning('Output directory {} already exists.'.format(
                    self._args.export_model))
                quit()
        if self._args.suffix:
            if any(i in self._args.suffix for i in [' ', '|', ':', ';', ',']):
                logger.error('Special character or space found '
                             'in provided suffix')
        tdict = app_reader(self._args.rbh, self._args.target - 1,
                           self._args.template - 1)
        trans_genes_dict = model_loader(self._model,
                                        self._args.ignore_na, tdict,
                                        self._args.suffix)
        homolo_reactions = list()
        for r in self._model.reactions:
            if r.id in self._model.model:
                if trans_genes_dict[r][2]:
                    r.properties['genes'] = remove_gap(
                        str(trans_genes_dict[r][1]))
                    if self._args.suffix:
                        r.properties['id'] += '_{}'.format(self._args.suffix)
                        for cpd, v in r.equation.compounds:
                            cpd._compartment += '_{}'.format(self._args.suffix)
                    homolo_reactions.append(r)
        if self._args.export_model is None:
            with open(self._args.output + '.yaml', 'w') as o:
                ModelWriter().write_reactions(o, homolo_reactions)
        elif self._args.export_model is not None:
            mkdir('{}'.format(self._args.export_model))
            reaction_list = [i.id for i in homolo_reactions]
            original_reactions = [i.id for i in self._model.reactions]
            for reaction in original_reactions:
                self._model.reactions.discard(reaction)
            for reaction in homolo_reactions:
                self._model.reactions.add_entry(reaction)
            compound_set = set()
            for reaction in self._model.reactions:
                for compound in reaction.equation.compounds:
                    compound_set.add(compound[0].name)
            compound_removal_set = set()
            for compound in self._model.compounds:
                if compound.id not in compound_set:
                    compound_removal_set.add(compound.id)
            for cpd in compound_removal_set:
                self._model.compounds.discard(cpd)
            exchange_del_set = set()
            for key in self._model.exchange:
                if key.name not in compound_set:
                    exchange_del_set.add(key)
            for ex in exchange_del_set:
                del self._model.exchange[ex]
            lim_del_set = set()
            for key in self._model.limits:
                if key not in reaction_list:
                    lim_del_set.add(key)
            for lim in lim_del_set:
                del self._model.exchange[lim]
            write_yaml_model(self._model, dest='{}'.format(
                self._args.export_model), split_subsystem=False)


def app_reader(app_file, query, template):
    '''This function will read in a gene mapping file and produce a
    gene to gene dictionary.

    The input is the app_file argument that the user enters, the query genome,
    which is the number of the genome that the user wishes to make a model of,
    and the number of the template genome which the user wishes to use to
    create the new model from.
    '''
    trans_dict = defaultdict(list)
    # Check what csv.reader in Python 3 takes (byte string or unicode string)
    for x, row in enumerate(csv.reader(app_file, delimiter='\t')):
        temp_l = []
        quer_l = []
        temp = row[template]
        quer = row[query]
        if temp != '-':
            if ',' in temp:
                temp_split = temp.split(',')
            else:
                temp_split = [temp]
            for i in temp_split:
                temp_l.append(i.strip())
            if ',' in quer:
                quer_split = quer.split(',')
            else:
                quer_split = [quer]
            for i in quer_split:
                quer_l.append(i.strip())
        for i in temp_l:
            for j in quer_l:
                trans_dict[i].append(j)
    return trans_dict


def model_loader(nm, ignore_na, translation_dict, suffix=None):
    '''This function will translate genes in one model
    based on a gene to gene mapping dictionary.

    The function takes a native model object, a true or false
    ignore_na argument, and a gene mapping dictionary translation_dict
    that is made by the app_reader function. The ignore_na argument
    will determine if reactions with no genes will be kept in the final
    translated model or not. This function will then translate the gene
    associations based on the gene mapping dictionary and produce a
    new set of gene associations based on the mapped associations. The
    gene expressions are evaluated based on the logic to see if they should
    be included in the new model.
    '''
    new_translation_dict = {}
    translated_genes = {}
    target_genes_l = {}
    target_model_reactions = []
    translation_dict.pop('-', None)
    for key, value in translation_dict.items():
        if len(value) > 1:
            value_s = '({})'.format(' or '.join(value))
        else:
            value_s, = value
        new_translation_dict[key] = value_s
        for i in value:
            target_genes_l[i] = True
    mm = nm.create_metabolic_model()
    model_rxns = [i for i in mm.reactions]
    print('ReactionID\tOriginal_Genes\tTranslated_Genes\tIn_final_model')
    for entry in nm.reactions:
        if entry.id in model_rxns:
            if entry.genes is None:
                target_model_reactions.append(entry.id)
                translated_genes[entry] = [None, None,
                                           not ignore_na]
                id = entry.id
                if suffix is not None:
                    id = id + '_{}'.format(suffix)
                print(u'{}\t{}\t{}\t{}'.format(
                    id, entry.genes, 'None',
                    not ignore_na))
            elif entry.genes is not None:
                genes = re.sub(r'\?', '', entry.genes)
                e = boolean.Expression(genes)
                e_1 = e.substitute(
                    lambda v: new_translation_dict.get(
                        v.symbol, '-') != '-')
                genes_1 = entry.genes
                gene_list = get_gene_list(genes)
                for g in gene_list:
                    if g in new_translation_dict:
                        genes = re.sub(r'\b' + g + r'\b',
                                       new_translation_dict[g], genes)
                    else:
                        genes = re.sub(r'\b' + g + r'\b',
                                       '-', genes)
                id = entry.id
                if suffix is not None:
                    id = id + '_{}'.format(suffix)
                print(u'{}\t{}\t{}\t{}'.format(
                    id, entry.genes, genes, e_1.value))
                translated_genes[entry] = [genes_1, genes, e_1.value]
    return translated_genes


def get_gene_list(genes):
    '''This function is used to get a list of genes from a gene expression
    '''
    gene_list = re.sub(r'[,()\']*', '', genes)
    gene_list = re.sub(r'\band\b', '', gene_list)
    gene_list = re.sub(r'\bor\b', '', gene_list)
    gene_list = re.split(r'\s+', gene_list)  # Transfer string to list
    gene_list = frozenset(gene_list)
    return gene_list


def remove_gap(string):
    '''This function is used to simplify translated gene expression strings.
    '''
    while True:
        prev = string
        string = re.sub(r'or - or', 'or', string)
        string = re.sub(r' or -', '', string)
        string = re.sub(r'\(-\)', '-', string)
        string = re.sub(r'- or ', '', string)
        string = re.sub(r'\(\)', '', string)
        string = re.sub(r'\'\'', '', string)
        if prev == string:
            return string



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>


import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm.datasource.reaction import parse_reaction
from psamm.expression import boolean
from psamm.lpsolver import generic
from psamm import fluxanalysis
from psamm.commands import gimme
from psamm.util import MaybeRelative


class TestAddReactions(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('A[e] => B[e]'))
        self._database.set_reaction('rxn_2', parse_reaction('B[e] => C[e]'))
        self._database.set_reaction('rxn_3', parse_reaction('B[e] <=> D[e]'))
        self._database.set_reaction('rxn_4', parse_reaction('C[e] <=> E[e]'))
        self._database.set_reaction('rxn_5', parse_reaction('D[e] => E[e]'))
        self._database.set_reaction('rxn_6', parse_reaction('E[e] =>'))
        self._database.set_reaction('rxn_7', parse_reaction('F[e] =>'))
        self._database.set_reaction('ex_A', parse_reaction('A[e] <=>'))

        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)
        self._assoc = {
            'rxn_1': str('gene_1'),
            'rxn_2': str('gene_2'),
            'rxn_3': str('gene_5'),
            'rxn_4': str('gene_3 or gene_4'),
            'rxn_5': str('gene_5 and gene_6'),
            'rxn_7': str('gene5 and gene4')
        }
        self._obj_reaction = 'rxn_6'
        self.empty_dict = {}

        try:
            self._solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')
        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Needs and integer programming solver')

    def test_reverse_model(self):
        test_associations = {
            'rxn_1': str('gene_1'),
            'rxn_2': str('gene_2'),
            'rxn_3_forward': str('gene_5'),
            'rxn_3_reverse': str('gene_5'),
            'rxn_4_forward': str('gene_3 or gene_4'),
            'rxn_4_reverse': str('gene_3 or gene_4'),
            'rxn_5': str('gene_5 and gene_6')
        }
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            self._mm.make_irreversible(self._assoc,
                                       exclude_list=['ex_A', 'rxn_6', 'rxn_7'])
        self.assertEqual(split_rxns, set([('rxn_4_forward', 'rxn_4_reverse'),
                                          ('rxn_3_forward', 'rxn_3_reverse')]))
        self.assertEqual(reversible_gene_assoc, test_associations)
        self.assertEqual(set([i for i in mm_irreversible.reactions]),
                         {'rxn_1', 'rxn_2', 'rxn_3_forward', 'rxn_3_reverse',
                          'rxn_4_forward', 'rxn_4_reverse', 'rxn_5',
                          'rxn_6', 'rxn_7', 'ex_A'})

    def test_revese_model_limits(self):
        mm = self._mm.copy()
        mm.limits['rxn_3'].lower = 0
        empty_dict = {}
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            mm.make_irreversible(self._assoc,
                                 exclude_list=['ex_A', 'rxn_6', 'rxn_7'])
        self.assertEqual(
            mm_irreversible.limits['rxn_3_forward'].bounds, (0, 1000))
        self.assertEqual(
            mm_irreversible.limits['rxn_3_reverse'].bounds, (0, 0))
        mm.limits['rxn_3'].lower = -1000
        mm.limits['rxn_3'].upper = 0
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            mm.make_irreversible(self._assoc,
                                 exclude_list=['ex_A', 'rxn_6', 'rxn_7'])
        self.assertEqual(
            mm_irreversible.limits['rxn_3_forward'].bounds, (0, 0))
        self.assertEqual(
            mm_irreversible.limits['rxn_3_reverse'].bounds, (0, 1000))

    def test_all_reverse_model(self):
        test_associations = {
            'rxn_1_forward': str('gene_1'),
            'rxn_1_reverse': str('gene_1'),
            'rxn_3_forward': str('gene_5'),
            'rxn_3_reverse': str('gene_5'),
            'rxn_4_forward': str('gene_3 or gene_4'),
            'rxn_4_reverse': str('gene_3 or gene_4'),
        }
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            self._mm.make_irreversible(
                self._assoc,
                exclude_list=['ex_A', 'rxn_2', 'rxn_5', 'rxn_6', 'rxn_7'],
                all_reversible=True)
        self.assertEqual(split_rxns, set([('rxn_1_forward', 'rxn_1_reverse'),
                                          ('rxn_4_forward', 'rxn_4_reverse'),
                                          ('rxn_3_forward', 'rxn_3_reverse')]))
        self.assertEqual(reversible_gene_assoc, test_associations)
        self.assertEqual(set([i for i in mm_irreversible.reactions]),
                         {'rxn_1_forward', 'rxn_1_reverse', 'rxn_2',
                          'rxn_3_forward', 'rxn_3_reverse',
                          'rxn_4_forward', 'rxn_4_reverse', 'rxn_5',
                          'rxn_6', 'rxn_7', 'ex_A'})

    def test_parse_transcriptome_file(self):
        f = ['gene\texpression', 'gene_1\t15', 'gene_2\t20',
             'gene_3\t15', 'gene_4\t25', 'gene_5\t10', 'gene_6\t25']
        d = gimme.parse_transcriptome_file(f, 20)
        self.assertEqual(d, {'gene_1': 5.0, 'gene_3': 5.0, 'gene_5': 10.0})

    def test_get_rxn_value(self):
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict =\
            self._mm.make_irreversible(
                self._assoc,
                exclude_list=['ex_A', 'rxn_6'])
        f = ['gene\texpression', 'gene_1\t15', 'gene_2\t20',
             'gene_3\t15', 'gene_4\t10', 'gene_5\t10', 'gene_6\t25']
        d = gimme.parse_transcriptome_file(f, 20)
        root_single = gimme.get_rxn_value(boolean.Expression(
            reversible_gene_assoc['rxn_1'])._root, d)
        self.assertEqual(root_single, 5)

        root_and = gimme.get_rxn_value(boolean.Expression(
            reversible_gene_assoc['rxn_5'])._root, d)
        self.assertEqual(root_and, 10)

        root_or = gimme.get_rxn_value(boolean.Expression(
            reversible_gene_assoc['rxn_4_forward'])._root, d)
        self.assertEqual(root_or, 5)

        root_none = gimme.get_rxn_value(boolean.Expression(
            reversible_gene_assoc['rxn_2'])._root, d)
        self.assertEqual(root_none, None)

    def test_gimme_model(self):
        f = ['gene\texpression', 'gene_1\t15', 'gene_2\t20',
             'gene_3\t15', 'gene_4\t25', 'gene_5\t10', 'gene_6\t25']
        threshold_dict = gimme.parse_transcriptome_file(f, 20)
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            self._mm.make_irreversible(
                self._assoc,
                exclude_list=['ex_A', 'rxn_6'])
        p = fluxanalysis.FluxBalanceProblem(mm_irreversible, generic.Solver())
        final_model, used_exchange, below_threshold_ids, incon_score = \
            gimme.solve_gimme_problem(
                p, mm_irreversible, 'rxn_6',
                reversible_gene_assoc, split_rxns, threshold_dict,
                MaybeRelative(20))
        self.assertEqual(incon_score, 100)
        self.assertEqual(final_model, set(['rxn_1', 'rxn_2', 'rxn_4']))
        self.assertEqual(below_threshold_ids, set(['rxn_1', 'rxn_3', 'rxn_5']))

    def test_gimme_model_relative(self):
        f = ['gene\texpression', 'gene_1\t15', 'gene_2\t20',
             'gene_3\t15', 'gene_4\t30', 'gene_5\t10', 'gene_6\t25']
        threshold_dict = gimme.parse_transcriptome_file(f, 20)
        mm_irreversible, reversible_gene_assoc, split_rxns, self.empty_dict = \
            self._mm.make_irreversible(
                self._assoc,
                exclude_list=['ex_A', 'rxn_6'])
        p = fluxanalysis.FluxBalanceProblem(mm_irreversible, generic.Solver())
        final_model, used_exchange, below_threshold_ids, incon_score = \
            gimme.solve_gimme_problem(
                p, mm_irreversible, 'rxn_6',
                reversible_gene_assoc, split_rxns, threshold_dict,
                MaybeRelative('2%'))
        self.assertEqual(incon_score, 100)
        self.assertEqual(final_model, set(['rxn_1', 'rxn_2', 'rxn_4']))
        self.assertEqual(below_threshold_ids, set(['rxn_1', 'rxn_3', 'rxn_5']))
